# places.py
#
# Copyright 2020 Martin Abente Lahaye
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from gi.repository import Adw, GLib, Gtk, GObject

from . import utils
from . import logger
from .place import PortfolioPlace
from .devices import PortfolioDevices
from .translation import gettext as _


class PortfolioPlaces(Gtk.Stack):
    __gtype_name__ = "PortfolioPlaces"

    __gsignals__ = {
        "updated": (GObject.SignalFlags.RUN_LAST, None, (str,)),
        "removing": (GObject.SignalFlags.RUN_LAST, None, (str,)),
        "removed": (GObject.SignalFlags.RUN_LAST, None, (str, bool)),
        "failed": (GObject.SignalFlags.RUN_LAST, None, (str,)),
        "unlock": (GObject.SignalFlags.RUN_LAST, None, (object,)),
    }

    FLATPAK_INFO = os.path.join(os.path.abspath(os.sep), ".flatpak-info")
    PORTFOLIO_SYSTEM_DIR = os.path.abspath(os.sep)
    PORTFOLIO_SYSTEM_DIR_FLATPAK = os.path.join(os.path.abspath(os.sep), "run", "host")
    PORTFOLIO_HOME_DIR = os.environ.get("PORTFOLIO_HOME_DIR", os.path.expanduser("~"))

    XDG_DOWNLOAD = utils.get_path_for_xdg(GLib.UserDirectory.DIRECTORY_DOWNLOAD)
    XDG_DOCUMENTS = utils.get_path_for_xdg(GLib.UserDirectory.DIRECTORY_DOCUMENTS)
    XDG_PICTURES = utils.get_path_for_xdg(GLib.UserDirectory.DIRECTORY_PICTURES)
    XDG_MUSIC = utils.get_path_for_xdg(GLib.UserDirectory.DIRECTORY_MUSIC)
    XDG_VIDEOS = utils.get_path_for_xdg(GLib.UserDirectory.DIRECTORY_VIDEOS)

    XDG_TRASH = "Trash"
    XDG_TRASH_NAME = _("Trash")

    HOST_PERMISSION = ["host"]
    HOME_PERMISSION = ["host", "home"]
    DOWNLOAD_PERMISSION = ["host", "home", "xdg-download"]
    DOCUMENTS_PERMISSION = ["host", "home", "xdg-documents"]
    PICTURES_PERMISSION = ["host", "home", "xdg-pictures"]
    MUSIC_PERMISSION = ["host", "home", "xdg-music"]
    VIDEOS_PERMISSION = ["host", "home", "xdg-videos"]
    TRASH_PERMISSION = ["host", "home"]

    def __init__(self, **kargs):
        super().__init__(**kargs)
        self._setup()

    def _setup(self):
        self.props.visible = True
        self.props.transition_type = Gtk.StackTransitionType.CROSSFADE

        self._permissions = None

        self._devices = PortfolioDevices()
        self._devices.connect("added", self._on_device_added)
        self._devices.connect("removed", self._on_device_removed)
        self._devices.connect("encrypted-added", self._on_encrypted_added)

        # begin UI structure

        self._groups_box = Gtk.Box()
        self._groups_box.props.visible = True
        self._groups_box.props.orientation = Gtk.Orientation.VERTICAL

        self._message_box = Gtk.Box()
        self._message_box.props.visible = True
        self._message_box.props.orientation = Gtk.Orientation.VERTICAL

        self._places_group = Adw.PreferencesGroup()
        self._places_group.props.title = _("Places")
        self._places_group.props.visible = True

        self._devices_group = Adw.PreferencesGroup()
        self._devices_group.props.title = _("Devices")
        self._devices_group.props.visible = True
        self._devices_group.get_style_context().add_class("devices-group")

        # places

        if self._has_permission_for(self.HOME_PERMISSION):
            self._add_place(
                self._places_group,
                "user-home-symbolic",
                _("Home"),
                self.PORTFOLIO_HOME_DIR,
            )
        if self._has_permission_for(self.DOWNLOAD_PERMISSION) and self.XDG_DOWNLOAD:
            self._add_place(
                self._places_group,
                "folder-download-symbolic",
                os.path.basename(self.XDG_DOWNLOAD),
                self.XDG_DOWNLOAD,
            )
        if self._has_permission_for(self.DOCUMENTS_PERMISSION) and self.XDG_DOCUMENTS:
            self._add_place(
                self._places_group,
                "folder-documents-symbolic",
                os.path.basename(self.XDG_DOCUMENTS),
                self.XDG_DOCUMENTS,
            )
        if self._has_permission_for(self.PICTURES_PERMISSION) and self.XDG_PICTURES:
            self._add_place(
                self._places_group,
                "folder-pictures-symbolic",
                os.path.basename(self.XDG_PICTURES),
                self.XDG_PICTURES,
            )
        if self._has_permission_for(self.MUSIC_PERMISSION) and self.XDG_MUSIC:
            self._add_place(
                self._places_group,
                "folder-music-symbolic",
                os.path.basename(self.XDG_MUSIC),
                self.XDG_MUSIC,
            )
        if self._has_permission_for(self.VIDEOS_PERMISSION) and self.XDG_VIDEOS:
            self._add_place(
                self._places_group,
                "folder-videos-symbolic",
                os.path.basename(self.XDG_VIDEOS),
                self.XDG_VIDEOS,
            )
        if self._has_permission_for(self.TRASH_PERMISSION):
            self._add_place(
                self._places_group,
                "user-trash-symbolic",
                self.XDG_TRASH_NAME,
                self.XDG_TRASH,
            )

        # static devices

        if self._has_permission_for(self.HOST_PERMISSION):
            self._add_place(
                self._devices_group,
                "drive-harddisk-ieee1394-symbolic",
                _("System"),
                self.PORTFOLIO_SYSTEM_DIR,
            )

        if self._has_permission_for(self.HOST_PERMISSION) and self._is_flatpak():
            self._add_place(
                self._devices_group,
                "drive-harddisk-ieee1394-symbolic",
                _("Host"),
                self.PORTFOLIO_SYSTEM_DIR_FLATPAK,
            )

        self._groups_box.append(self._places_group)
        self._groups_box.append(self._devices_group)

        self._places_listbox = utils.find_child_by_id(self._places_group, "listbox")
        self._devices_listbox = utils.find_child_by_id(self._devices_group, "listbox")

        # no places message

        message = Gtk.Label()
        message.props.visible = True
        message.props.label = _("No places found")
        message.get_style_context().add_class("no-places")

        self._message_box.append(message)

        # finalize UI structure

        self.add_named(self._groups_box, "groups")
        self.add_named(self._message_box, "message")

        # dynamic devices

        self._devices.scan()

        # update visibility

        self._update_visibility()

    def _update_visibility(self):
        self._update_stack_visibility()
        self._update_places_group_visibility()
        self._update_device_group_visibility()

    def _update_stack_visibility(self):
        groups = len(list(self._places_listbox))
        devices = len(list(self._devices_listbox))

        if not groups and not devices:
            self.set_visible_child_name("message")
        else:
            self.set_visible_child_name("groups")

    def _update_places_group_visibility(self):
        visible = len(list(self._places_listbox)) >= 1
        self._places_group.props.visible = visible

    def _update_device_group_visibility(self):
        visible = len(list(self._devices_listbox)) >= 1
        self._devices_group.props.visible = visible

    def _get_permissions(self):
        if self._permissions is not None:
            return self._permissions

        info = GLib.KeyFile()
        info.load_from_file(self.FLATPAK_INFO, GLib.KeyFileFlags.NONE)
        permissions = info.get_value("Context", "filesystems")

        if permissions is not None:
            self._permissions = set(permissions.split(";"))
        else:
            self._permissions = set()

        return self._permissions

    def _is_flatpak(self):
        return os.path.exists(self.FLATPAK_INFO)

    def _has_permission_for(self, required):
        # not using flatpak, so access to all
        if not self._is_flatpak():
            return True

        permissions = self._get_permissions()
        required = set(required)
        negated = set([f"!{r}" for r in required])

        if required.intersection(permissions):
            return True
        if negated.intersection(permissions):
            return False

        return False

    def _add_place(self, group, icon, name, path):
        place = PortfolioPlace()
        place.set_custom_icon_name(icon)
        place.set_title(name)
        place.set_subtitle_sately(path)
        place.path = path
        place.props.activatable = True
        place.connect("activated", self._on_place_activated)
        group.add(place)

        return place

    def _find_place_by_device_uuid(self, listbox, device):
        for place in listbox:
            if place.uuid == device.uuid:
                return place
        return None

    def _filter_device(self, device):
        if device.mount_point is None:
            logger.debug(f"No mount point for {device}")
            return False
        if not os.access(device.mount_point, os.R_OK):
            logger.debug(f"No permissions for {device}")
            return True
        if device.mount_point == "/" and self._is_flatpak():
            logger.debug(f"Skip root directory for {device}")
            return True

        return False

    def _update_place_from_device(self, place, device):
        if place is None:
            return

        if self._filter_device(device):
            self._devices_group.remove(place)
            return

        place.path = device.mount_point
        place.set_title(device.label)
        place.set_subtitle_sately(device.mount_point)

    def _on_place_activated(self, place):
        if place.path is not None:
            self.emit("updated", place.path)
        elif place.encrypted is not None:
            self.emit("unlock", place.encrypted)
        elif place.device is not None:
            place.device.mount(self._on_insert_finished)

    def _on_insert_finished(self, device, encrypted, success):
        logger.debug(f"insert finished {device}")
        if success:
            self.emit("updated", device.mount_point)

    def _on_encrypted_added(self, devices, encrypted):
        logger.debug(f"added {encrypted}")

        place = self._add_place(
            self._devices_group,
            "system-lock-screen-symbolic",
            encrypted.get_friendly_label(),
            None,
        )

        place.uuid = encrypted.uuid
        place.encrypted = encrypted

        place.eject.props.visible = True
        place.eject.connect("clicked", self._on_encrypted_eject, encrypted)

        self._update_visibility()

    def _on_device_added(self, devices, device):
        logger.debug(f"added {device}")

        if self._filter_device(device):
            return

        place = self._add_place(
            self._devices_group,
            "drive-removable-media-symbolic",
            device.label,
            device.mount_point,
        )

        place.uuid = device.uuid
        place.device = device
        place.eject.props.visible = True
        place.eject.connect("clicked", self._on_eject, device)

        device.connect("updated", self._on_device_updated)

        self._update_place_from_device(place, device)
        self._update_visibility()

    def _on_device_removed(self, devices, device):
        logger.debug(f"removed {device}")
        place = self._find_place_by_device_uuid(self._devices_listbox, device)

        if place is not None:
            self._devices_group.remove(place)

        self.emit("removed", device.mount_point, device.safely_removed)

        self._update_visibility()

    def _on_device_updated(self, device):
        logger.debug(f"updated {device}")
        place = self._find_place_by_device_uuid(self._devices_listbox, device)
        self._update_place_from_device(place, device)
        self._update_visibility()

    def _on_eject(self, button, device):
        logger.debug(f"eject {device}")
        self.emit("removing", device.mount_point)
        device.eject(self._on_eject_finished)

    def _on_eject_finished(self, device, success):
        logger.debug(f"eject finished {device} {success}")
        if success:
            self._on_device_removed(None, device)
        else:
            self.emit("failed", device.mount_point)

    def _on_encrypted_eject(self, button, encrypted):
        logger.debug(f"encrypted eject {encrypted}")
        encrypted.eject(self._on_encrypted_eject_finished)

    def _on_encrypted_eject_finished(self, encrypted, success):
        logger.debug(f"encrypted eject finished {encrypted}")
        if success:
            self._on_device_removed(None, encrypted)
        else:
            self.emit("failed", None)



# place.py
#
# Copyright 2021 Martin Abente Lahaye
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Adw, Gtk


@Gtk.Template(resource_path="/dev/tchx84/Portfolio/place.ui")
class PortfolioPlace(Adw.ActionRow):
    __gtype_name__ = "PortfolioPlace"

    icon = Gtk.Template.Child()
    eject = Gtk.Template.Child()
    mount = None
    path = ""
    uuid = ""
    encrypted = None
    device = None

    def set_custom_icon_name(self, name):
        self.icon.set_from_icon_name(name)

    def set_subtitle_sately(self, subtitle):
        if subtitle is not None:
            self.set_subtitle(subtitle)



# loading.py
#
# Copyright 2023 Martin Abente Lahaye
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk


@Gtk.Template(resource_path="/dev/tchx84/Portfolio/loading.ui")
class PortfolioLoading(Gtk.Box):
    __gtype_name__ = "PortfolioLoading"

    title = Gtk.Template.Child()
    progress = Gtk.Template.Child()
    description = Gtk.Template.Child()
    details = Gtk.Template.Child()

    def update(self, title=None, progress=None, description=None, details=None):
        if title is not None:
            self.title.set_text(title)
        if progress is not None:
            self.progress.set_fraction(progress)
        if description is not None:
            self.description.set_text(description)
        if details is not None:
            self.details.set_text(details)

        self.description.props.visible = True if self.description.props.label else False
        self.details.props.visible = True if self.details.props.label else False

    def pulse(self):
        self.progress.pulse()

    def clean(self):
        self.update("", 0.0, "", "")



# properties.py
#
# Copyright 2023 Martin Abente Lahaye
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GObject, Gtk


@Gtk.Template(resource_path="/dev/tchx84/Portfolio/properties.ui")
class PortfolioProperties(Gtk.Box):
    __gtype_name__ = "PortfolioProperties"

    property_name = Gtk.Template.Child()
    property_location = Gtk.Template.Child()
    property_type = Gtk.Template.Child()
    property_size = Gtk.Template.Child()
    property_created = Gtk.Template.Child()
    property_modified = Gtk.Template.Child()
    property_accessed = Gtk.Template.Child()
    property_permissions_owner = Gtk.Template.Child()
    property_permissions_group = Gtk.Template.Child()
    property_permissions_others = Gtk.Template.Child()
    property_owner = Gtk.Template.Child()
    property_group = Gtk.Template.Child()

    def __init__(self, worker, **kwargs):
        super().__init__(**kwargs)
        self._worker = worker
        self._setup()

    def _setup(self):
        self._worker.bind_property(
            "name",
            self.property_name,
            "label",
            GObject.BindingFlags.SYNC_CREATE,
        )
        self._worker.bind_property(
            "location",
            self.property_location,
            "label",
            GObject.BindingFlags.SYNC_CREATE,
        )
        self._worker.bind_property(
            "type",
            self.property_type,
            "label",
            GObject.BindingFlags.SYNC_CREATE,
        )
        self._worker.bind_property(
            "size",
            self.property_size,
            "label",
            GObject.BindingFlags.SYNC_CREATE,
        )
        self._worker.bind_property(
            "created",
            self.property_created,
            "label",
            GObject.BindingFlags.SYNC_CREATE,
        )
        self._worker.bind_property(
            "modified",
            self.property_modified,
            "label",
            GObject.BindingFlags.SYNC_CREATE,
        )
        self._worker.bind_property(
            "accessed",
            self.property_accessed,
            "label",
            GObject.BindingFlags.SYNC_CREATE,
        )
        self._worker.bind_property(
            "permissions_owner",
            self.property_permissions_owner,
            "label",
            GObject.BindingFlags.SYNC_CREATE,
        )
        self._worker.bind_property(
            "permissions_group",
            self.property_permissions_group,
            "label",
            GObject.BindingFlags.SYNC_CREATE,
        )
        self._worker.bind_property(
            "permissions_others",
            self.property_permissions_others,
            "label",
            GObject.BindingFlags.SYNC_CREATE,
        )
        self._worker.bind_property(
            "owner",
            self.property_owner,
            "label",
            GObject.BindingFlags.SYNC_CREATE,
        )
        self._worker.bind_property(
            "group",
            self.property_group,
            "label",
            GObject.BindingFlags.SYNC_CREATE,
        )



# window.py
#
# Copyright 2020 Martin Abente Lahaye
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from .translation import gettext as _

from gi.repository import Adw, Gtk, GLib

from . import utils
from . import logger
from .popup import PortfolioPopup
from .worker import PortfolioCutWorker
from .worker import PortfolioCopyWorker
from .worker import PortfolioDeleteWorker
from .worker import PortfolioLoadWorker
from .worker import PortfolioOpenWorker
from .worker import PortfolioPropertiesWorker
from .worker import PortfolioRestoreTrashWorker
from .worker import PortfolioDeleteTrashWorker
from .worker import PortfolioSendTrashWorker
from .worker import PortfolioLoadTrashWorker
from .places import PortfolioPlaces
from .properties import PortfolioProperties
from .about import PortfolioAbout
from .passphrase import PortfolioPassphrase
from .placeholder import PortfolioPlaceholder
from .loading import PortfolioLoading
from .files import PortfolioFiles
from .menu import PortfolioMenu
from .settings import PortfolioSettings
from .trash import default_trash


@Gtk.Template(resource_path="/dev/tchx84/Portfolio/window.ui")
class PortfolioWindow(Adw.ApplicationWindow):
    __gtype_name__ = "PortfolioWindow"

    previous = Gtk.Template.Child()
    next = Gtk.Template.Child()
    search = Gtk.Template.Child()
    rename = Gtk.Template.Child()
    detail = Gtk.Template.Child()
    delete = Gtk.Template.Child()
    cut = Gtk.Template.Child()
    copy = Gtk.Template.Child()
    paste = Gtk.Template.Child()
    select_all = Gtk.Template.Child()
    select_none = Gtk.Template.Child()
    new_folder = Gtk.Template.Child()
    delete_trash = Gtk.Template.Child()
    restore_trash = Gtk.Template.Child()
    close_button = Gtk.Template.Child()
    stop_button = Gtk.Template.Child()
    go_top_button = Gtk.Template.Child()
    about_back_button = Gtk.Template.Child()
    properties_back_button = Gtk.Template.Child()

    search_box = Gtk.Template.Child()
    search_entry = Gtk.Template.Child()
    popup_box = Gtk.Template.Child()
    action_stack = Gtk.Template.Child()
    tools_stack = Gtk.Template.Child()
    navigation_box = Gtk.Template.Child()
    selection_box = Gtk.Template.Child()
    selection_tools = Gtk.Template.Child()
    navigation_tools = Gtk.Template.Child()
    places_box = Gtk.Template.Child()
    places_inner_box = Gtk.Template.Child()
    places_popup_box = Gtk.Template.Child()
    content_stack = Gtk.Template.Child()
    loading_box = Gtk.Template.Child()
    loading_inner_box = Gtk.Template.Child()
    content_box = Gtk.Template.Child()
    files_stack = Gtk.Template.Child()
    files_box = Gtk.Template.Child()
    files_title = Gtk.Template.Child()
    about_box = Gtk.Template.Child()
    about_inner_box = Gtk.Template.Child()
    close_box = Gtk.Template.Child()
    close_tools = Gtk.Template.Child()
    stop_box = Gtk.Template.Child()
    stop_tools = Gtk.Template.Child()
    trash_tools = Gtk.Template.Child()
    about_deck = Gtk.Template.Child()
    content_deck = Gtk.Template.Child()
    placeholder_box = Gtk.Template.Child()
    placeholder_inner_box = Gtk.Template.Child()
    menu_button = Gtk.Template.Child()
    home_menu_button = Gtk.Template.Child()
    content_inner_box = Gtk.Template.Child()
    go_top_revealer = Gtk.Template.Child()
    properties_box = Gtk.Template.Child()
    properties_inner_box = Gtk.Template.Child()
    passphrase_title = Gtk.Template.Child()
    passphrase_box = Gtk.Template.Child()
    passphrase_inner_box = Gtk.Template.Child()
    passphrase_back_button = Gtk.Template.Child()

    SEARCH_DELAY = 500
    LOAD_ANIMATION_DELAY = 250

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._setup()

    def _setup(self):
        self._popup = None
        self._places_popup = None
        self._worker = None
        self._busy = False
        self._to_copy = []
        self._to_cut = []
        self._force_go_home = False
        self._history = []
        self._index = -1
        self._search_delay_handler_id = 0
        self._load_delay_handler_id = 0

        self._settings = PortfolioSettings()
        self._settings.connect("notify::show-hidden", self._on_show_hidden_changed)
        self._settings.connect("notify::sort-order", self._on_sort_order_changed)

        self.previous.connect("clicked", self._on_go_previous)
        self.next.connect("clicked", self._on_go_next)
        self.rename.connect("clicked", self._on_rename_clicked)
        self.detail.connect("clicked", self._on_detail_clicked)
        self.delete.connect("clicked", self._on_delete_clicked)
        self.cut.connect("clicked", self._on_cut_clicked)
        self.copy.connect("clicked", self._on_copy_clicked)
        self.paste.connect("clicked", self._on_paste_clicked)
        self.restore_trash.connect("clicked", self._on_restore_trash_clicked)
        self.delete_trash.connect("clicked", self._on_delete_trash_clicked)
        self.select_all.connect("clicked", self._on_select_all)
        self.select_none.connect("clicked", self._on_select_none)
        self.new_folder.connect("clicked", self._on_new_folder)
        self.close_button.connect("clicked", self._on_button_closed)
        self.go_top_button.connect("clicked", self._go_to_top)
        self.stop_button.connect("clicked", self._on_stop_clicked)
        self.about_back_button.connect("clicked", self._on_about_back_clicked)
        self.properties_back_button.connect("clicked", self._on_properties_back_clicked)
        self.passphrase_back_button.connect("clicked", self._on_passphrase_back_clicked)

        # XXX no model for options yet so this...
        self.menu_popover = PortfolioMenu(self._settings)
        self.menu_popover.connect("show-about", self._on_about_clicked)
        self.menu_button.props.popover = self.menu_popover
        self.menu_button.connect("activate", self._on_menu_button_clicked)

        self.home_menu_popover = PortfolioMenu(self._settings)
        self.home_menu_popover.connect("show-about", self._on_about_clicked)
        self.home_menu_button.props.popover = self.home_menu_popover
        self.home_menu_button.connect("activate", self._on_menu_button_clicked)

        self.search.connect("toggled", self._on_search_toggled)
        self.search_entry.connect("search-changed", self._on_search_changed)
        self.search_entry.connect("stop-search", self._on_search_stopped)

        # XXX Compose these widgets directly in .ui files
        self.files = PortfolioFiles()
        self.files.connect("activated", self._on_files_activated)
        self.files.connect("selected", self._on_files_selected)
        self.files.connect("rename-started", self._on_files_rename_started)
        self.files.connect("rename-finished", self._on_files_rename_finished)
        self.files.connect("rename-failed", self._on_files_rename_failed)
        self.files.connect("add-failed", self._on_files_add_failed)
        self.files.connect("adjustment-changed", self._on_files_adjustment_changed)
        self.files.sort_order = self._settings.sort_order
        self.content_inner_box.append(self.files)

        places = PortfolioPlaces()
        places.connect("updated", self._on_places_updated)
        places.connect("removing", self._on_places_removing)
        places.connect("removed", self._on_places_removed)
        places.connect("failed", self._on_places_failed)
        places.connect("unlock", self._on_places_unlock)
        self.places_inner_box.append(places)

        self._properties_worker = PortfolioPropertiesWorker()
        self.properties_inner_box.append(PortfolioProperties(self._properties_worker))

        self.about_inner_box.append(PortfolioAbout())

        self.passphrase = PortfolioPassphrase()
        self.passphrase.connect("unlocked", self._on_places_unlocked)
        self.passphrase_inner_box.append(self.passphrase)

        self.placeholder_inner_box.append(PortfolioPlaceholder())

        self.loading = PortfolioLoading()
        self.loading_inner_box.append(self.loading)

        self.content_deck.connect("notify::visible-child", self._on_content_folded)
        self.connect("close-request", self._on_close_request)

    def _populate(self, directory):
        self.files.switch_to_navigation_mode()

        if self._worker is not None:
            self._worker.stop()

        if default_trash.is_trash(directory):
            loader_class = PortfolioLoadTrashWorker
        else:
            loader_class = PortfolioLoadWorker

        self._worker = loader_class(directory, self._settings.show_hidden)
        self._worker.connect("started", self._on_load_started)
        self._worker.connect("updated", self._on_load_updated)
        self._worker.connect("finished", self._on_load_finished)
        self._worker.connect("failed", self._on_load_failed)
        self._worker.start()

    def _paste(self, Worker, to_paste):
        directory = self._history[self._index]

        self._worker = Worker(to_paste, directory)
        self._worker.connect("started", self._on_paste_started)
        self._worker.connect("updated", self._on_paste_updated)
        self._worker.connect("post-update", self._on_paste_post_updated)
        self._worker.connect("finished", self._on_paste_finished)
        self._worker.connect("failed", self._on_paste_failed)
        self._worker.connect("stopped", self._on_paste_stopped)
        self._worker.start()

    def _paste_finish(self):
        self._busy = False
        self._clean_workers()
        self.loading.clean()

        self._to_cut = []
        self._to_copy = []

        self.files.unselect_all()
        self._update_all()

    def _delete_finish(self):
        self._busy = False
        self._clean_workers()
        self.loading.clean()

        self.files.unselect_all()
        self._update_all()

    def _get_row(self, model, treepath):
        return model.get_iter(treepath)

    def _go_to_top(self, *args):
        self.files.go_to_top()

    def _go_back_to_homepage(self):
        self.content_deck.set_visible_child(self.places_box)

    def _go_to_files_view(self, duration):
        self.content_deck.set_visible_child(self.files_stack)
        self.files_stack.props.transition_duration = duration
        self.files_stack.set_visible_child(self.files_box)

    def _move(self, path, navigating=False):
        self._clean_popups()

        if path is None:
            return
        elif default_trash.is_trash(path) or utils.is_file_dir(path):
            self._update_history(path, navigating)
            self._populate(path)
        else:
            self._open(path)

    def _open(self, path):
        self._worker = PortfolioOpenWorker(path)
        self._worker.connect("started", self._on_open_started)
        self._worker.connect("updated", self._on_open_updated)
        self._worker.connect("finished", self._on_open_finished)
        self._worker.connect("failed", self._on_open_failed)
        self._worker.start()

    def _reset_to_path(self, path):
        self._history = []
        self._index = -1
        self._move(path, False)

    def _refresh(self):
        if self._index > -1:
            self._move(self._history[self._index], True)

    def _notify(self, description, on_confirm, on_cancel, on_trash, autoclose, data):
        self._clean_popups()

        self._popup = PortfolioPopup(
            description, on_confirm, on_cancel, on_trash, autoclose, data
        )
        self.popup_box.append(self._popup)
        self._popup.props.reveal_child = True

    def _places_notify(self, description):
        self._clean_places_popups()

        self._places_popup = PortfolioPopup(description, None, None, None, False, None)
        self.places_popup_box.append(self._places_popup)
        self._places_popup.props.reveal_child = True

    def _clean_workers(self):
        del self._worker
        self._worker = None

    def _clean_popups(self):
        if self._popup is None:
            return
        if self._popup.get_parent() is not None:
            self.popup_box.remove(self._popup)
        self._popup = None

    def _clean_places_popups(self):
        if self._places_popup is None:
            return
        if self._places_popup.get_parent() is not None:
            self.places_popup_box.remove(self._places_popup)
        self._places_popup = None

    def _clean_loading_delay(self):
        if self._load_delay_handler_id != 0:
            GLib.Source.remove(self._load_delay_handler_id)
            self._load_delay_handler_id = 0

    def _update_history(self, path, navigating):
        if path not in self._history or not navigating:
            del self._history[self._index + 1 :]
            self._history.append(path)
            self._index += 1

    def _update_all(self):
        self._update_search()
        self._update_treeview()
        self._update_content_stack()
        self._update_navigation()
        self._update_navigation_tools()
        self._update_trash_tools()
        self._update_selection()
        self._update_selection_tools()
        self._update_action_stack()
        self._update_tools_stack()
        self._update_menu()

    def _update_search(self):
        sensitive = not self.files.is_editing and not self._busy
        self.search.props.sensitive = sensitive
        self.search_entry.sensitive = sensitive

    def _update_treeview(self):
        self.files.update(not self._busy)

    def _update_content_stack(self):
        if self._busy:
            return
        elif self.files.is_empty:
            self.content_stack.set_visible_child(self.placeholder_box)
        else:
            self.content_stack.set_visible_child(self.content_box)

    def _update_navigation(self):
        count = self.files.selected_count
        selected = count >= 1

        if selected or self._busy:
            self.previous.props.sensitive = False
            self.next.props.sensitive = False
            return

        self.previous.props.sensitive = True
        self.next.props.sensitive = (
            True if len(self._history) - 1 > self._index else False
        )

    def _update_selection(self):
        sensitive = not self.files.is_editing and not self._busy

        self.select_all.props.sensitive = sensitive
        self.select_none.props.sensitive = sensitive

    def _update_action_stack(self):
        count = self.files.selected_count
        selected = count >= 1
        child = self.selection_box if selected else self.navigation_box
        self.action_stack.set_visible_child(child)

    def _update_tools_stack(self):
        directory = self._history[self._index]
        if default_trash.in_trash(directory):
            self.tools_stack.set_visible_child(self.trash_tools)
            return

        count = self.files.selected_count
        selected = count >= 1
        child = self.selection_tools if selected else self.navigation_tools
        self.tools_stack.set_visible_child(child)

    def _update_selection_tools(self):
        count = self.files.selected_count
        sensitive = count >= 1 and not self.files.is_editing and not self._busy

        self.delete.props.sensitive = sensitive
        self.cut.props.sensitive = sensitive
        self.copy.props.sensitive = sensitive

        self._update_rename()
        self._update_detail()

    def _update_navigation_tools(self):
        count = self.files.selected_count
        selected = count >= 1
        to_paste = len(self._to_cut) >= 1 or len(self._to_copy) >= 1
        self.paste.props.sensitive = not selected and to_paste and not self._busy
        self.new_folder.props.sensitive = not selected and not self._busy

    def _update_trash_tools(self):
        selected = self.files.selected_count >= 1
        is_trash = default_trash.is_trash(self._history[self._index])
        self.restore_trash.props.sensitive = selected and is_trash
        self.delete_trash.props.sensitive = selected and is_trash

    def _update_rename(self):
        count = self.files.selected_count
        sensitive = count == 1 and not self.files.is_editing and not self._busy
        self.rename.props.sensitive = sensitive

    def _update_detail(self):
        count = self.files.selected_count
        sensitive = count == 1 and not self.files.is_editing and not self._busy
        self.detail.props.sensitive = sensitive

    def _update_directory_title(self):
        directory = self._history[self._index]

        if default_trash.is_trash(directory):
            name = PortfolioPlaces.XDG_TRASH_NAME
        else:
            name = os.path.basename(directory)

        self.files_title.props.title = name

    def _update_filter(self):
        self.files.filter = self.search_entry.get_text()
        self._update_content_stack()

        self._search_delay_handler_id = 0
        return GLib.SOURCE_REMOVE

    def _update_menu(self):
        self.menu_popover.is_sensitive = not self._busy
        self.home_menu_popover.is_sensitive = not self._busy

    def _reset_search(self):
        self.search.set_active(False)
        self.search_entry.set_text("")
        self.search.grab_focus()

    def _on_files_activated(self, files, path):
        self._move(path)

    def _on_files_selected(self, files):
        if self._busy is True:
            return
        self._update_all()

    def _on_files_rename_started(self, files):
        self._update_search()
        self._update_selection()
        self._update_selection_tools()

    def _on_files_rename_finished(self, files):
        self._update_all()

    def _on_files_rename_failed(self, files, new_name):
        self._on_rename_clicked(None)
        self._notify(
            _("%s already exists") % new_name,
            None,
            self._on_popup_closed,
            None,
            True,
            None,
        )

    def _on_files_add_failed(self, files):
        self._notify(
            _("No permissions on this directory"),
            None,
            self._on_popup_closed,
            None,
            True,
            None,
        )

    def _on_files_adjustment_changed(self, files, reveal):
        self.go_top_revealer.props.reveal_child = reveal

    def _on_open_started(self, worker):
        self._busy = True
        self.loading.update(_("Opening"), 0.0, "", "")
        self.content_stack.set_visible_child(self.loading_box)

        self._update_all()

    def _on_open_updated(self, worker):
        self.loading.pulse()

    def _on_open_finished(self, worker):
        self._busy = False
        self._clean_workers()
        self._update_all()

    def _on_open_failed(self, worker, path):
        self._busy = False
        self._clean_workers()

        name = os.path.basename(path)
        self.loading.update(description=_("Could not open %s") % name)

        self.action_stack.set_visible_child(self.close_box)
        self.tools_stack.set_visible_child(self.close_tools)

    def _on_load_started(self, worker, directory):
        self._busy = True

        self.files.clear()
        self._update_directory_title()
        self._reset_search()
        self._update_all()

        self._load_delay_handler_id = GLib.timeout_add(
            self.LOAD_ANIMATION_DELAY,
            self._on_load_started_delayed,
        )

    def _on_load_started_delayed(self):
        self.loading.update(_("Loading"), 0.0, "", "")
        self.content_stack.set_visible_child(self.loading_box)

        self._update_all()

        self._load_delay_handler_id = 0
        return GLib.SOURCE_REMOVE

    def _on_load_updated(self, worker, directory, found, index, total):
        for name, path, icon in found:
            self.files.add_row(icon, name, path)

        self.loading.update(progress=(index + 1) / total)

    def _on_load_finished(self, worker, directory):
        self._busy = False
        self._clean_workers()
        self._clean_loading_delay()

        self._update_all()
        self.files.update_scrolling()

    def _on_load_failed(self, worker, directory):
        self._busy = False
        self._clean_workers()
        self._clean_loading_delay()

        name = os.path.basename(directory)
        self.loading.update(
            title=_("Loading"), description=_("Could not load %s") % name
        )
        self.content_stack.set_visible_child(self.loading_box)

        self.files._clear_select_and_go()
        self._force_go_home = True
        self.action_stack.set_visible_child(self.close_box)
        self.tools_stack.set_visible_child(self.close_tools)

    def _on_go_previous(self, button):
        if self._index == 0:
            self._go_back_to_homepage()
        else:
            self.files.to_go_to_path = self._history[self._index]
            self._index -= 1
            self._move(self._history[self._index], True)

    def _on_go_next(self, button):
        self._index += 1
        self._move(self._history[self._index], True)

    def _on_search_toggled(self, button):
        toggled = self.search.get_active()
        self.search_box.props.search_mode_enabled = toggled

    def _on_search_changed(self, entry):
        if self._search_delay_handler_id != 0:
            GLib.Source.remove(self._search_delay_handler_id)
        self._search_delay_handler_id = GLib.timeout_add(
            self.SEARCH_DELAY, self._update_filter
        )

    def _on_search_stopped(self, entry):
        self._reset_search()

    def _on_detail_clicked(self, button):
        path = self.files.get_selected_path()
        self.show_properties(path)

    def _on_rename_clicked(self, button):
        self.files.rename_selected_row()

    def _on_delete_clicked(self, button):
        selection = self.files.get_selection()
        count = len(selection)
        directory = self._history[self._index]

        if count == 1:
            name = os.path.basename(selection[0][0])
        else:
            name = _("these %d files") % count

        description = _("Delete %s permanently?") % name

        self._notify(
            description,
            self._on_delete_confirmed,
            self._on_popup_closed,
            self._on_trash_instead if default_trash.has_trash(directory) else None,
            False,
            selection,
        )

    def _on_cut_clicked(self, button):
        selection = self.files.get_selection()
        count = len(selection)

        self._to_cut = selection
        self._to_copy = []

        if count == 1:
            name = os.path.basename(selection[0][0])
            description = _("%s will be moved") % name
        else:
            description = _("%d files will be moved") % count

        self._notify(description, None, None, None, True, None)

        self.files.unselect_all()

    def _on_copy_clicked(self, button):
        selection = self.files.get_selection()
        count = len(selection)

        self._to_copy = selection
        self._to_cut = []

        if count == 1:
            name = os.path.basename(selection[0][0])
            description = _("%s will be copied") % name
        else:
            description = _("%d files will be copied") % count

        self._notify(description, None, None, None, True, None)

        self.files.unselect_all()

    def _on_paste_clicked(self, button):
        to_paste = self._to_copy if self._to_copy else self._to_cut
        Worker = PortfolioCopyWorker if self._to_copy else PortfolioCutWorker

        directory = self._history[self._index]
        should_warn = any(
            [
                os.path.dirname(path) != directory
                and os.path.lexists(os.path.join(directory, os.path.basename(path)))
                for path, ref in to_paste
            ]
        )

        if not should_warn:
            self._paste(Worker, to_paste)
            return

        self._notify(
            _("Files will be overwritten, proceed?"),
            self._on_paste_confirmed,
            self._on_popup_closed,
            None,
            False,
            (to_paste, Worker),
        )

    def _on_paste_confirmed(self, button, popup, data):
        to_paste, Worker = data
        self._clean_popups()
        self._paste(Worker, to_paste)

    def _on_paste_started(self, worker, total):
        self._busy = True
        self.loading.update(_("Pasting"), 0.0, "", "")
        self.content_stack.set_visible_child(self.loading_box)

        self._update_all()

        self.action_stack.set_visible_child(self.stop_box)
        self.tools_stack.set_visible_child(self.stop_tools)

    def _on_paste_post_updated(self, worker, name, path, icon, overwritten):
        if overwritten:
            return
        if not os.path.exists(path):
            logger.debug(f"Attempting to add unexisting {path}")
            return

        self.files.add_row(icon, name, path)

    def _on_paste_updated(self, worker, path, index, total, current_bytes, total_bytes):
        description = os.path.basename(path)
        human_current_bytes = utils.get_size_for_humans(current_bytes)
        human_total_bytes = utils.get_size_for_humans(total_bytes)
        self.loading.update(
            description=description,
            progress=index / total,
            details=_("%s of %s") % (human_current_bytes, human_total_bytes),
        )

    def _on_paste_finished(self, worker, total):
        self._paste_finish()

    def _on_paste_failed(self, worker, path):
        self._busy = False
        self._clean_workers()
        self.loading.clean()

        self._to_cut = []
        self._to_copy = []

        name = os.path.basename(path)
        self.loading.update(description=_("Could not paste %s") % name)

        self.action_stack.set_visible_child(self.close_box)
        self.tools_stack.set_visible_child(self.close_tools)

    def _on_paste_stopped(self, worker):
        self._paste_finish()
        # XXX nuclear fix for when parent directory does not get to be updated
        self._refresh()

    def _on_trash_instead(self, button, popup, selection):
        self._clean_popups()

        # clean history entries from deleted paths
        directory = self._history[self._index]
        self._history = [
            path
            for path in self._history
            if not path.startswith(directory) or path == directory
        ]

        self._worker = PortfolioSendTrashWorker(selection)
        self._worker.connect("started", self._on_delete_started)
        self._worker.connect("pre-update", self._on_delete_pre_updated)
        self._worker.connect("updated", self._on_delete_updated)
        self._worker.connect("finished", self._on_delete_finished)
        self._worker.connect("failed", self._on_delete_failed)
        self._worker.connect("stopped", self._on_delete_stopped)
        self._worker.start()

    def _on_delete_confirmed(self, button, popup, selection):
        self._clean_popups()

        # clean history entries from deleted paths
        directory = self._history[self._index]
        self._history = [
            path
            for path in self._history
            if not path.startswith(directory) or path == directory
        ]

        self._worker = PortfolioDeleteWorker(selection)
        self._worker.connect("started", self._on_delete_started)
        self._worker.connect("pre-update", self._on_delete_pre_updated)
        self._worker.connect("updated", self._on_delete_updated)
        self._worker.connect("finished", self._on_delete_finished)
        self._worker.connect("failed", self._on_delete_failed)
        self._worker.connect("stopped", self._on_delete_stopped)
        self._worker.start()

    def _on_delete_started(self, worker):
        self._busy = True
        self.loading.update(_("Deleting"), 0.0, "", "")
        self.content_stack.set_visible_child(self.loading_box)

        self._update_all()

        self.action_stack.set_visible_child(self.stop_box)
        self.tools_stack.set_visible_child(self.stop_tools)

    def _on_delete_pre_updated(self, worker, path):
        name = os.path.basename(path)
        self.loading.update(description=name)

    def _on_delete_updated(self, worker, path, row, index, total):
        self.files.remove_row(row)
        self.loading.update(progress=(index + 1) / total)

    def _on_delete_finished(self, worker, total):
        self._delete_finish()

    def _on_delete_failed(self, worker, path):
        self._busy = False
        self._clean_workers()
        self.loading.clean()

        name = os.path.basename(path)
        self.loading.update(description=_("Could not delete %s") % name)

        self.action_stack.set_visible_child(self.close_box)
        self.tools_stack.set_visible_child(self.close_tools)

    def _on_delete_stopped(self, worker):
        self._delete_finish()

    def _on_popup_closed(self, button, popup, data):
        self._clean_popups()

    def _on_button_closed(self, button):
        self.loading.clean()
        self.files.unselect_all()
        self._update_all()

        if self._force_go_home is False:
            return

        self._force_go_home = False
        self._go_back_to_homepage()

    def _on_stop_clicked(self, button):
        self.loading.update(title=_("Stopping"))
        self._worker.stop()

    def _on_select_all(self, button):
        self.files.select_all()

    def _on_select_none(self, button):
        self.files.unselect_all()

    def _on_new_folder(self, button):
        directory = self._history[self._index]
        self.files.add_new_folder_row(directory)

    def _on_restore_trash_clicked(self, button):
        selection = self.files.get_selection()
        paths = [default_trash.get_orig_path(path) for path, ref in selection]

        overwrites = any([os.path.lexists(path) for path in paths if path])
        duplicates = len(set(paths)) != len(paths)

        if not overwrites and not duplicates:
            self._on_restore_trash_confirmed(None, None, selection)
            return

        self._notify(
            _("Files will be overwritten, proceed?"),
            self._on_restore_trash_confirmed,
            self._on_popup_closed,
            None,
            False,
            selection,
        )

    def _on_restore_trash_confirmed(self, button, popup, selection):
        self._clean_popups()

        self._worker = PortfolioRestoreTrashWorker(selection)
        self._worker.connect("started", self._on_restore_trash_started)
        self._worker.connect("pre-update", self._on_restore_trash_pre_updated)
        self._worker.connect("updated", self._on_restore_trash_updated)
        self._worker.connect("finished", self._on_restore_trash_finished)
        self._worker.connect("failed", self._on_restore_trash_failed)
        self._worker.connect("stopped", self._on_restore_trash_stopped)
        self._worker.start()

    def _on_restore_trash_started(self, worker):
        self._busy = True

        self.loading.update(_("Restoring"), 0.0, "", "")
        self.content_stack.set_visible_child(self.loading_box)

        self._update_all()

        self.action_stack.set_visible_child(self.stop_box)
        self.tools_stack.set_visible_child(self.stop_tools)

    def _on_restore_trash_pre_updated(self, worker, path):
        name = os.path.basename(path)
        self.loading.update(description=name)

    def _on_restore_trash_updated(self, worker, path, row, index, total):
        self.files.remove_row(row)
        self.loading.update(progress=(index + 1) / total)

    def _on_restore_trash_finished(self, worker, total):
        self._busy = False
        self._clean_workers()
        self.loading.clean()

        self.files.unselect_all()
        self._update_all()

    def _on_restore_trash_failed(self, worker, path):
        self._busy = False
        self._clean_workers()
        self.loading.clean()

        self.loading.update(description=_("Could not restore %s") % path)

        self.action_stack.set_visible_child(self.close_box)
        self.tools_stack.set_visible_child(self.close_tools)

    def _on_restore_trash_stopped(self, worker):
        self._delete_finish()

    def _on_delete_trash_clicked(self, button):
        selection = self.files.get_selection()
        count = len(selection)

        if count == 1:
            name = os.path.basename(selection[0][0])
        else:
            name = _("these %d files") % count

        description = _("Delete %s permanently?") % name

        self._notify(
            description,
            self._on_delete_trash_confirmed,
            self._on_popup_closed,
            None,
            False,
            selection,
        )

    def _on_delete_trash_confirmed(self, button, popup, selection):
        self._clean_popups()

        self._worker = PortfolioDeleteTrashWorker(selection)
        self._worker.connect("started", self._on_delete_trash_started)
        self._worker.connect("pre-update", self._on_delete_trash_pre_updated)
        self._worker.connect("updated", self._on_delete_trash_updated)
        self._worker.connect("finished", self._on_delete_trash_finished)
        self._worker.connect("failed", self._on_delete_trash_failed)
        self._worker.connect("stopped", self._on_delete_trash_stopped)
        self._worker.start()

    def _on_delete_trash_started(self, worker):
        self._busy = True

        self.loading.update(_("Deleting"), 0.0, "", "")
        self.content_stack.set_visible_child(self.loading_box)

        self._update_all()

        self.action_stack.set_visible_child(self.stop_box)
        self.tools_stack.set_visible_child(self.stop_tools)

    def _on_delete_trash_pre_updated(self, worker, path):
        name = os.path.basename(path)
        self.loading.update(description=name)

    def _on_delete_trash_updated(self, worker, path, row, index, total):
        self.files.remove_row(row)
        self.loading.update(progress=(index + 1) / total)

    def _on_delete_trash_finished(self, worker, total):
        self._delete_finish()

    def _on_delete_trash_failed(self, worker, path):
        self._busy = False
        self._clean_workers()
        self.loading.clean()

        self.loading.update(description=_("Could not delete %s") % path)

        self.action_stack.set_visible_child(self.close_box)
        self.tools_stack.set_visible_child(self.close_tools)

    def _on_delete_trash_stopped(self, worker):
        self._delete_finish()

    def _on_places_updated(self, button, path):
        self._reset_to_path(path)
        self._go_to_files_view(duration=0)

    def _on_places_removing(self, button, path):
        self._places_notify(_("Removing device, please wait"))

    def _on_places_removed(self, button, path, safely):
        if safely is True:
            self._places_notify(_("Device can be removed"))

        if self._index == -1:
            return

        directory = self._history[self._index]
        if not path or not directory.startswith(path):
            return

        self._go_back_to_homepage()

    def _on_places_failed(self, button, path):
        self._places_notify(_("Device is busy, can't be removed"))

    def _on_places_unlock(self, button, encrypted):
        self.passphrase.unlock(encrypted)
        self.passphrase_title.props.title = encrypted.get_friendly_label()
        self.content_deck.set_visible_child(self.files_stack)
        self.files_stack.set_visible_child(self.passphrase_box)

    def _on_places_unlocked(self, widget, path):
        self._reset_to_path(path)
        self._go_to_files_view(duration=200)

    def _on_passphrase_back_clicked(self, button):
        self.passphrase.clean()
        self.content_deck.set_visible_child(self.places_box)

    def _on_about_clicked(self, button):
        self.about_deck.set_visible_child(self.about_box)

    def _on_about_back_clicked(self, button):
        self.about_deck.set_visible_child(self.content_deck)

    def _on_properties_back_clicked(self, button):
        self.content_deck.set_visible_child(self.files_stack)

    def _on_show_hidden_changed(self, settings, data):
        self._refresh()

    def _on_sort_order_changed(self, settings, data):
        self.files.sort_order = self._settings.sort_order
        self._refresh()

    def _on_menu_button_clicked(self, button):
        button.props.popover.popup()

    def _on_content_folded(self, deck, data=None):
        child = self.content_deck.get_visible_child()
        if child == self.places_box and self._worker is not None:
            self._worker.stop()
            self._clean_workers()
        elif child == self.files_box:
            self._properties_worker.stop()

    def _on_close_request(self, window):
        if self._worker is not None:
            self._worker.stop()
        self._properties_worker.stop()

    def open(self, path=PortfolioPlaces.PORTFOLIO_HOME_DIR, force_page_switch=False):
        path = utils.get_uri_path(path)

        # make sure it exists though !
        if not os.path.exists(path):
            logger.warning(_("Could not open %s") % path)
            return

        # if it's a file then use its parent folder
        if not os.path.isdir(path):
            self.files.to_select_path = path
            path = os.path.dirname(path)

        # XXX no support for background workers yet
        if self._busy and not isinstance(self._worker, PortfolioLoadWorker):
            logger.warning(_("Could not open %s") % path)
            return

        if isinstance(self._worker, PortfolioLoadWorker):
            self._worker.stop()
            self._clean_workers()
            self._clean_loading_delay()

        self._reset_to_path(path)

        if force_page_switch is False:
            return

        self.about_deck.set_visible_child(self.content_deck)
        self.content_deck.set_visible_child(self.files_stack)

    def show_properties(self, path, force_page_switch=False):
        self._properties_worker.props.path = path

        if force_page_switch is True:
            self.open(path, force_page_switch)

        self.content_deck.set_visible_child(self.properties_box)



# cache.py
#
# Copyright 2021 Martin Abente Lahaye
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GObject

from . import logger


class PortfolioCache(GObject.GObject):
    __gtype_name__ = "PortfolioCache"

    def __init__(self):
        super().__init__()
        self.deactivate()

    def has(self, key):
        if not self._active:
            return False

        return key in self._cache

    def retrieve(self, key):
        if not self.has(key):
            return None

        return self._cache[key]

    def store(self, key, value):
        if not self._active:
            return

        self._cache[key] = value

    def activate(self):
        logger.debug("cache activated")
        self._active = True
        self._cache = {}

    def deactivate(self):
        logger.debug("cache deactivated")
        self._active = False
        self._cache = {}


class cached(object):
    def __init__(self, function):
        self._function = function

    def __call__(self, *args, **kwargs):
        key = (self._function.__name__, args)

        if default_cache.has(key):
            return default_cache.retrieve(key)

        value = self._function(*args, **kwargs)
        default_cache.store(key, value)

        return value


default_cache = PortfolioCache()



# popup.py
#
# Copyright 2020 Martin Abente Lahaye
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib, Gtk


DEFAULT_CLOSE_TIME = 3


@Gtk.Template(resource_path="/dev/tchx84/Portfolio/popup.ui")
class PortfolioPopup(Gtk.Revealer):
    __gtype_name__ = "PortfolioPopup"

    description = Gtk.Template.Child()
    confirm_button = Gtk.Template.Child()
    cancel_button = Gtk.Template.Child()
    trash_button = Gtk.Template.Child()

    def __init__(self, description, on_confirm, on_cancel, on_trash, autoclose, data):
        super().__init__()

        self.description.set_text(description)

        if on_confirm is not None:
            self.confirm_button.connect("clicked", on_confirm, self, data)
        else:
            self.confirm_button.props.visible = False

        if on_cancel is not None:
            self.cancel_button.connect("clicked", on_cancel, self, data)
        else:
            self.cancel_button.connect("clicked", self._on_default_callback, self, data)

        if on_trash is not None:
            self.trash_button.connect("clicked", on_trash, self, data)
        else:
            self.trash_button.props.visible = False

        if autoclose is True:
            GLib.timeout_add_seconds(
                DEFAULT_CLOSE_TIME, self._on_default_callback, None, None, None
            )

    def set_description(self, description):
        self.description.set_text(description)

    def _on_default_callback(self, button, popup, data):
        if parent := self.get_parent():
            parent.remove(self)



# menu.py
#
# Copyright 2023 Martin Abente Lahaye
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Gio, GObject

from .settings import PortfolioSettings


@Gtk.Template(resource_path="/dev/tchx84/Portfolio/menu.ui")
class PortfolioMenu(Gtk.Popover):
    __gtype_name__ = "PortfolioMenu"

    __gsignals__ = {
        "show-about": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    menu_box = Gtk.Template.Child()
    show_hidden_button = Gtk.Template.Child()
    a_to_z_button = Gtk.Template.Child()
    last_modified_button = Gtk.Template.Child()
    help_button = Gtk.Template.Child()
    about_button = Gtk.Template.Child()

    def __init__(self, settings):
        super().__init__()
        self._setup(settings)

    @property
    def is_sensitive(self):
        return self.menu_box.props.sensitive

    @is_sensitive.setter
    def is_sensitive(self, value):
        self.menu_box.props.sensitive = value

    def _setup(self, settings):
        self._settings = settings
        self._settings.connect("notify::show-hidden", self._on_show_hidden_changed)
        self._settings.connect("notify::sort-order", self._on_sort_order_changed)

        self._show_hidden_handler_id = self.show_hidden_button.connect(
            "toggled", self._on_show_hidden_toggled
        )
        self._sort_order_handler_id = self.a_to_z_button.connect(
            "toggled", self._on_sort_by_toggled
        )

        self.help_button.connect("clicked", self._on_help_clicked)
        self.about_button.connect("clicked", self._on_about_clicked)

        self._on_show_hidden_changed(self._settings, None)
        self._on_sort_order_changed(self._settings, None)

    def _on_show_hidden_changed(self, settings, data):
        GObject.signal_handler_block(
            self.show_hidden_button,
            self._show_hidden_handler_id,
        )

        self.show_hidden_button.props.active = self._settings.show_hidden

        GObject.signal_handler_unblock(
            self.show_hidden_button,
            self._show_hidden_handler_id,
        )

    def _on_sort_order_changed(self, settings, data):
        GObject.signal_handler_block(
            self.a_to_z_button,
            self._sort_order_handler_id,
        )

        if self._settings.sort_order == PortfolioSettings.ALPHABETICAL_ORDER:
            self.a_to_z_button.props.active = True
        else:
            self.last_modified_button.props.active = True

        GObject.signal_handler_unblock(
            self.a_to_z_button,
            self._sort_order_handler_id,
        )

    def _on_show_hidden_toggled(self, button):
        self._settings.show_hidden = self.show_hidden_button.props.active

    def _on_sort_by_toggled(self, button):
        if self.a_to_z_button.props.active:
            self._settings.sort_order = PortfolioSettings.ALPHABETICAL_ORDER
        else:
            self._settings.sort_order = PortfolioSettings.MODIFIED_TIME_ORDER

    def _on_help_clicked(self, button):
        Gio.AppInfo.launch_default_for_uri("https://github.com/tchx84/Portfolio", None)

    def _on_about_clicked(self, button):
        self.emit("show-about")



# service.py
#
# Copyright 2021 Martin Abente Lahaye
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from gi.repository import Gio


class PortfolioService:
    def __init__(self, app):
        self._app = app
        self._dbus_id = None
        self._name_id = Gio.bus_own_name(
            Gio.BusType.SESSION,
            os.environ.get("PORTFOLIO_SERVICE_NAME", "org.freedesktop.FileManager1"),
            Gio.BusNameOwnerFlags.NONE,
            self._on_bus_acquired,
            self._on_name_acquired,
            self._on_name_lost,
        )

    def __del__(self):
        self.shutdown()

    def _on_bus_acquired(self, connection, name):
        xml = (
            Gio.resources_lookup_data(
                "/dev/tchx84/Portfolio/org.freedesktop.FileManager1.xml",
                Gio.ResourceLookupFlags.NONE,
            )
            .get_data()
            .decode("utf-8")
        )
        info = Gio.DBusNodeInfo.new_for_xml(xml)

        activation_id = connection.register_object(
            "/org/freedesktop/FileManager1", info.interfaces[0], self._on_called
        )

        assert activation_id > 0

    def _on_name_acquired(self, connection, name):
        pass

    def _on_name_lost(self, connection, name):
        pass

    def _on_called(self, connection, sender, path, iface, method, params, invocation):
        paths, _ = params

        if len(paths) == 0:
            invocation.return_value(None)
            return

        last_path = paths[-1]

        if method == "ShowItemProperties":
            _method = self._app.show_properties
        elif method in ["ShowFolders", "ShowItems"]:
            _method = self._app.open_path
        else:
            raise ValueError(f"{method} not supported")

        # at this point we know we can handle it, so just respond
        invocation.return_value(None)
        _method(last_path)

    def shutdown(self):
        Gio.bus_unown_name(self._name_id)



# files.py
#
# Copyright 2023 Martin Abente Lahaye
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from pathlib import Path

from gi.repository import Gtk, GLib, GObject

from . import utils
from . import logger
from .settings import PortfolioSettings
from .translation import gettext as _


@Gtk.Template(resource_path="/dev/tchx84/Portfolio/files.ui")
class PortfolioFiles(Gtk.ScrolledWindow):
    __gtype_name__ = "PortfolioFiles"

    __gsignals__ = {
        "selected": (GObject.SignalFlags.RUN_LAST, None, ()),
        "activated": (GObject.SignalFlags.RUN_LAST, None, (str,)),
        "rename-started": (GObject.SignalFlags.RUN_LAST, None, ()),
        "rename-finished": (GObject.SignalFlags.RUN_LAST, None, ()),
        "rename-failed": (GObject.SignalFlags.RUN_LAST, None, (str,)),
        "add-failed": (GObject.SignalFlags.RUN_LAST, None, ()),
        "adjustment-changed": (GObject.SignalFlags.RUN_LAST, None, (bool,)),
    }

    treeview = Gtk.Template.Child()
    name_column = Gtk.Template.Child()
    name_cell = Gtk.Template.Child()
    sorted = Gtk.Template.Child()
    filtered = Gtk.Template.Child()
    selection = Gtk.Template.Child()
    liststore = Gtk.Template.Child()

    ICON_COLUMN = 0
    NAME_COLUMN = 1
    PATH_COLUMN = 2

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._setup()

    def _setup(self):
        self._is_editing = False
        self._to_select_path = None
        self._to_select_row = None
        self._to_go_to_path = None
        self._to_go_to_row = None
        self._last_clicked = None
        self._dont_activate = False
        self._force_select = False
        self._last_vscroll_value = None
        self._filter = ""
        self._sort_order = PortfolioSettings.ALPHABETICAL_ORDER

        self.filtered.set_visible_func(self._filter_func, None)
        self.sorted.set_default_sort_func(self._sort_func, None)
        self.selection.connect("changed", self._on_selection_changed)
        self.selection.set_select_function(self._on_select)
        self.treeview.connect("row-activated", self._on_row_activated)

        self.name_cell.connect("editing-started", self._on_rename_started)
        self.name_cell.connect("edited", self._on_rename_updated)
        self.name_cell.connect("editing-canceled", self._on_rename_finished)

        self._adjustment = self.get_vadjustment()
        self._adjustment.connect("value-changed", self._on_adjustment_changed)

        self.clicks = Gtk.GestureClick.new()
        self.clicks.connect("pressed", self._on_clicked)
        self.treeview.add_controller(self.clicks)

        self.gestures = Gtk.GestureLongPress.new()
        self.gestures.connect("pressed", self._on_long_pressed)
        self.treeview.add_controller(self.gestures)

    @property
    def filter(self):
        return self._filter

    @filter.setter
    def filter(self, value):
        self._filter = value
        self.filtered.refilter()

    @property
    def to_select_path(self):
        return self._to_select_path

    @to_select_path.setter
    def to_select_path(self, value):
        self._to_select_path = value

    @property
    def to_go_to_path(self):
        return self._to_go_to_path

    @to_go_to_path.setter
    def to_go_to_path(self, value):
        self._to_go_to_path = value

    @property
    def sort_order(self):
        return self._sort_order

    @sort_order.setter
    def sort_order(self, value):
        self._sort_order = value

    @property
    def selected_count(self):
        return self.selection.count_selected_rows()

    @property
    def is_editing(self):
        return self._is_editing

    @property
    def is_empty(self):
        return len(self.sorted) == 0

    def _filter_func(self, model, row, data=None):
        path = model[row][self.PATH_COLUMN]
        if not self._filter:
            return True
        return self._filter.lower() in os.path.basename(path).lower()

    def _sort_func(self, model, row1, row2, data=None):
        path1 = model[row1][self.PATH_COLUMN]
        path2 = model[row2][self.PATH_COLUMN]

        row1_is_dir = utils.is_file_dir(path1)
        row2_is_dir = utils.is_file_dir(path2)

        if row1_is_dir and not row2_is_dir:
            return -1
        elif not row1_is_dir and row2_is_dir:
            return 1

        if self._sort_order == PortfolioSettings.ALPHABETICAL_ORDER:
            return self._sort_by_a_to_z(path1, path2)
        else:
            return self._sort_by_last_modified(path1, path2)

    def _sort_by_last_modified(self, path1, path2):
        st_mtime1 = utils.get_file_mtime(path1)
        st_mtime2 = utils.get_file_mtime(path2)

        if st_mtime1 < st_mtime2:
            return 1
        elif st_mtime1 > st_mtime2:
            return -1

        return 0

    def _sort_by_a_to_z(self, path1, path2):
        path1 = path1.lower()
        path2 = path2.lower()

        if path1 < path2:
            return -1
        elif path1 > path2:
            return 1

        return 0

    def _get_path(self, model, treepath):
        return model[model.get_iter(treepath)][self.PATH_COLUMN]

    def _go_to_selection(self):
        model, treepaths = self.selection.get_selected_rows()
        treepath = treepaths[-1]
        self.treeview.set_cursor_on_cell(
            treepath, self.name_column, self.name_cell, False
        )
        self.treeview.scroll_to_cell(treepath, None, False, 0, 0)

    def _go_to(self, row):
        result, row = self.filtered.convert_child_iter_to_iter(row)
        result, row = self.sorted.convert_child_iter_to_iter(row)

        treepath = self.sorted.get_path(row)

        self.treeview.scroll_to_cell(treepath, None, False, 0, 0)
        self._clear_to_go_to()

    def _select_and_go(self, row, edit=False):
        self.switch_to_selection_mode()

        result, row = self.filtered.convert_child_iter_to_iter(row)
        result, row = self.sorted.convert_child_iter_to_iter(row)

        self._select_row(row)
        GLib.idle_add(self._go_to_selection)

        if edit is True:
            GLib.timeout_add(100, self._wait_and_edit)

        self._clear_select_and_go()

    def _wait_and_edit(self):
        value = self._adjustment.get_value()

        if value == self._last_vscroll_value:
            self.rename_selected_row()
            self._last_vscroll_value = None
            return False

        self._last_vscroll_value = value
        return True

    def _clear_select_and_go(self):
        self._to_select_path = None
        self._to_select_row = None

    def _clear_to_go_to(self):
        self._to_go_to_path = None
        self._to_go_to_row = None

    def _on_selection_changed(self, selection):
        self._update_mode()
        self.emit("selected")

    def _on_select(self, selection, model, treepath, selected, data=None):
        should_select = False

        if self._force_select is True:
            should_select = True
        elif treepath != self._last_clicked and selected:
            should_select = False
        elif treepath != self._last_clicked and not selected:
            should_select = False
        elif treepath == self._last_clicked and not selected:
            should_select = True
        elif treepath == self._last_clicked and selected:
            should_select = True

        if treepath == self._last_clicked:
            self._last_clicked = None
            self._dont_activate = True

        return should_select

    def _on_row_activated(self, treeview, treepath, treecolumn, data=None):
        if self._dont_activate is True:
            self._dont_activate = False
            return
        if self.selection.get_mode() == Gtk.SelectionMode.NONE:
            path = self._get_path(self.sorted, treepath)
            self.emit("activated", path)

    def _on_clicked(self, gesture, n_press, x, y):
        result = self.treeview.get_path_at_pos(x, y)
        if result is None:
            return
        treepath, column, x, y = result
        self._last_clicked = treepath

    def _on_rename_started(self, cell_name, treepath, data=None):
        self._is_editing = True
        self.emit("rename-started")

    def _on_rename_updated(self, cell_name, treepath, new_name, data=None):
        old_path = self._get_path(self.sorted, treepath)
        directory = os.path.dirname(old_path)
        new_path = os.path.join(directory, new_name)

        if new_path == old_path:
            self._on_rename_finished()
            return

        try:
            # respect empty folders
            if os.path.lexists(new_path):
                raise FileExistsError(_("%s already exists") % new_path)

            os.rename(old_path, new_path)

            _treepath = Gtk.TreePath.new_from_string(treepath)
            _treepath = self.sorted.convert_path_to_child_path(_treepath)
            _treepath = self.filtered.convert_path_to_child_path(_treepath)

            row = self.liststore.get_iter(_treepath)
            self.liststore.set_value(row, self.PATH_COLUMN, new_path)
            self.liststore.set_value(row, self.NAME_COLUMN, new_name)
        except Exception as e:
            logger.debug(e)
            self.emit("rename-failed", new_name)
            return

        # take the user to the new position
        self._on_rename_finished()
        self._go_to_selection()

    def _on_rename_finished(self, *args):
        self.name_cell.props.editable = False
        self._is_editing = False
        self.emit("rename-finished")

    def _on_long_pressed(self, gesture, x, y):
        if self.selection.get_mode() == Gtk.SelectionMode.MULTIPLE:
            return

        self.switch_to_selection_mode()
        path = self.treeview.get_path_at_pos(x, y)

        if path is None:
            self.switch_to_navigation_mode()
            return

        treepath = path[0]
        self.selection.select_path(treepath)

        # because of the custom selection rules, is not guaranteed
        # that this will actually be selected so always update mode.
        self._update_mode()

    def _on_adjustment_changed(self, adjustment):
        alloc = self.get_allocation()
        reveal = (
            self._adjustment.get_value() > (alloc.height / 2) and not self._is_editing
        )
        self.emit("adjustment-changed", reveal)

    def _update_mode(self):
        count = self.selection.count_selected_rows()
        if count == 0:
            self.switch_to_navigation_mode()

    def _select_row(self, row):
        self._force_select = True
        self.selection.select_iter(row)
        self._force_select = False

    def select_all(self):
        self._force_select = True
        self.selection.select_all()
        self._force_select = False

    def unselect_all(self):
        self._force_select = True
        self.selection.unselect_all()
        self._force_select = False

    def switch_to_navigation_mode(self):
        self.selection.set_mode(Gtk.SelectionMode.NONE)

    def switch_to_selection_mode(self):
        self.selection.set_mode(Gtk.SelectionMode.MULTIPLE)

    def update(self, sensitive):
        self.treeview.props.sensitive = sensitive

    def update_scrolling(self):
        if self._to_select_row is not None:
            self._select_and_go(self._to_select_row)
        elif self._to_go_to_row is not None:
            self._go_to(self._to_go_to_row)
        else:
            self.go_to_top()

    def go_to_top(self):
        if len(self.sorted) >= 1:
            self.treeview.scroll_to_cell(0, None, True, 0, 0)

    def get_selection(self):
        model, treepaths = self.selection.get_selected_rows()
        selection = [
            (
                model[treepath][self.PATH_COLUMN],
                Gtk.TreeRowReference.new(model, treepath),
            )
            for treepath in treepaths
        ]
        return selection

    def get_selected_path(self):
        model, treepaths = self.selection.get_selected_rows()
        treepath = treepaths[-1]
        path = model[treepath][self.PATH_COLUMN]
        return path

    def add_row(self, icon, name, path):
        row = self.liststore.append([icon, name, path])

        if self._to_select_path == path:
            self._to_select_row = row

        if self._to_go_to_path == path:
            self._to_go_to_row = row

    def add_new_folder_row(self, directory):
        folder_name = utils.find_new_name(directory, _("New Folder"))
        path = os.path.join(directory, folder_name)

        try:
            Path(path).mkdir(parents=False, exist_ok=True)
        except Exception as e:
            logger.debug(e)
            self.emit("add-failed")
            return

        icon = utils.get_file_icon(path)
        row = self.liststore.append([icon, folder_name, path])
        self._select_and_go(row, edit=True)

    def remove_row(self, row):
        if row is None or not row.valid():
            return

        treepath = row.get_path()
        treepath = self.sorted.convert_path_to_child_path(treepath)
        treepath = self.filtered.convert_path_to_child_path(treepath)

        self.liststore.remove(self.liststore.get_iter(treepath))

    def rename_selected_row(self):
        self.name_cell.props.editable = True
        model, treepaths = self.selection.get_selected_rows()
        treepath = treepaths[-1]
        self.treeview.set_cursor_on_cell(
            treepath, self.name_column, self.name_cell, True
        )

    def clear(self):
        self.liststore.clear()



# passphrase.py
#
# Copyright 2023 Martin Abente Lahaye
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from gi.repository import Gtk, GObject

from . import logger
from .translation import gettext as _


@Gtk.Template(resource_path="/dev/tchx84/Portfolio/passphrase.ui")
class PortfolioPassphrase(Gtk.Box):
    __gtype_name__ = "PortfolioPassphrase"

    __gsignals__ = {
        "unlocked": (GObject.SignalFlags.RUN_LAST, None, (str,)),
    }

    passphrase_entry = Gtk.Template.Child()
    passphrase_label = Gtk.Template.Child()
    passphrase_spinner = Gtk.Template.Child()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._setup()

    def _setup(self):
        self._encrypted = None
        self.passphrase_entry.connect("activate", self._on_passphrase_activate)

    def _on_passphrase_activate(self, button):
        self.passphrase_label.props.visible = False
        self.passphrase_spinner.props.visible = True
        self.passphrase_spinner.props.spinning = True

        passphrase = self.passphrase_entry.get_text()
        self._encrypted.unlock(passphrase, self._on_places_unlock_finished)

    def _on_places_unlock_finished(self, device, encrypted, success):
        self.clean()

        if device is None:
            logger.debug(f"Failed to unlock {encrypted}")
        elif not os.access(device.mount_point, os.R_OK):
            logger.debug(f"No permissions for {device}")
        elif success is True:
            self._encrypted = None
            self.emit("unlocked", device.mount_point)
            return

        self._encrypted = encrypted
        self.passphrase_entry.grab_focus()
        self.passphrase_label.set_text(_("Sorry, that didn't work"))

    def clean(self):
        self.passphrase_entry.set_text("")
        self.passphrase_label.set_text("")
        self.passphrase_label.props.visible = True
        self.passphrase_spinner.props.visible = False
        self.passphrase_spinner.props.spinning = False

    def unlock(self, encrypted):
        self.clean()
        self._encrypted = encrypted
        self.passphrase_entry.grab_focus()



# utils.py
#
# Copyright 2020 Martin Abente Lahaye
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import re

from gi.repository import GLib, Gio

from .cache import cached


def find_new_name(directory, name, fmt="%s(%d)"):
    counter = 1

    while os.path.lexists(os.path.join(directory, name)):
        components = re.split(r"(\(\d+\)$)", name)
        if len(components) > 1:
            name = "".join(components[:-2])

        name = fmt % (name, counter)
        counter += 1

    return name


def count(path):
    _count = 1

    if os.path.isdir(path):
        for directory, dirs, files in os.walk(path):
            _count += len(files)

    return _count


def flatten_walk(path):
    _paths = []

    def _callback(error):
        raise error

    if os.path.isdir(path):
        for directory, dirs, files in os.walk(path, topdown=False, onerror=_callback):
            for file in files:
                _paths += [os.path.join(directory, file)]
            _paths += [directory]
    else:
        _paths += [path]

    return _paths


def get_uri_path(string):
    try:
        uri = GLib.uri_parse(string, GLib.UriFlags.NONE)
        return uri.get_path()
    except:
        return string


def get_path_for_xdg(directory):
    path = GLib.get_user_special_dir(directory)

    if path is None:
        return None
    if path == GLib.get_home_dir():
        return None
    if not GLib.file_test(path, GLib.FileTest.EXISTS):
        return None
    if not GLib.file_test(path, GLib.FileTest.IS_DIR):
        return None

    return path


@cached
def get_file_mtime(string):
    return os.lstat(string).st_mtime


@cached
def is_file_dir(string):
    return os.path.isdir(string)


def get_file_icon(path):
    return (
        Gio.file_new_for_path(path)
        .query_info(
            Gio.FILE_ATTRIBUTE_STANDARD_ICON,
            Gio.FileQueryInfoFlags.NOFOLLOW_SYMLINKS,
        )
        .get_icon()
    )


def is_flatpak():
    return os.path.exists(os.path.join(os.path.sep, ".flatpak-info"))


def sync_folder(path):
    fd = os.open(path, os.O_DIRECTORY)
    os.fsync(fd)
    os.close(fd)


def find_mount_point(path):
    path = os.path.abspath(path)
    while not os.path.ismount(path):
        path = os.path.dirname(path)
    return path


# https://gist.github.com/cbwar/d2dfbc19b140bd599daccbe0fe925597
def get_size_for_humans(num):
    for unit in ["", "k", "M", "G", "T", "P", "E", "Z"]:
        if abs(num) < 1000.0:
            return "%3.1f %sB" % (num, unit)
        num /= 1000.0
    return "%.1f%sB" % (num, "Y")


def find_child_by_id(container, child_id):
    for widget in list(container):
        if widget.get_buildable_id() == child_id:
            return widget
        if child := find_child_by_id(widget, child_id):
            return child
    return None



# trash.py
#
# Copyright 2021 Martin Abente Lahaye
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import shutil

from datetime import datetime
from urllib.parse import quote, unquote
from configparser import ConfigParser

from gi.repository import GLib, Gio, GObject

from . import utils
from . import logger


class PortfolioTrash(GObject.GObject):
    # https://specifications.freedesktop.org/trash-spec/trashspec-1.0.html

    __gtype_name__ = "PortfolioTrash"

    def __init__(self):
        super().__init__()

        self._manager = Gio.VolumeMonitor.get()
        self._manager.connect("mount-added", self._on_mount_changed)
        self._manager.connect("mount-removed", self._on_mount_changed)

        self.setup()

    def _setup_trash_dir(self, trash_dir):
        os.makedirs(os.path.join(trash_dir, "info"), exist_ok=True)
        os.makedirs(os.path.join(trash_dir, "files"), exist_ok=True)

    def _on_mount_changed(self, monitor, mount):
        self.setup()

    def setup(self):
        self._trash = {}

        mounts = self.get_devices_trash() + [self.get_home_trash()]
        for mount_point, trash_dir in mounts:
            self._trash[mount_point] = trash_dir

    def is_trash(self, path):
        return path == "Trash"

    def in_trash(self, path):
        if self.is_trash(path):
            return True

        for _, trash_dir in self._trash.items():
            if path.startswith(trash_dir):
                return True

        return False

    def has_trash(self, path):
        mount_point = utils.find_mount_point(path)
        return mount_point in self._trash

    def get_home_trash(self):
        if "PORTFOLIO_XDG_DATA_DIRS" in os.environ:
            data_dir = os.environ.get("PORTFOLIO_XDG_DATA_DIRS")
        elif utils.is_flatpak() and "HOST_XDG_DATA_DIRS" in os.environ:
            data_dir = os.environ.get("HOST_XDG_DATA_DIRS")
        elif not utils.is_flatpak() and GLib.get_user_data_dir():
            data_dir = GLib.get_user_data_dir()
        else:
            data_dir = os.path.join(os.path.expanduser("~"), ".local", "share")

        self._home_trash = os.path.join(data_dir, "Trash")
        mount_point = utils.find_mount_point(data_dir)

        return mount_point, self._home_trash

    def get_devices_trash(self):
        trash = []

        for mount in self._manager.get_mounts():
            mount_point = mount.get_root().get_path()
            if not mount_point or not os.access(mount_point, os.W_OK):
                logger.debug("Can't find mount_point, skipping...")
                continue

            user_id = os.getuid()
            if not user_id:
                logger.debug("Can't find user_id, skipping...")
                continue

            path = os.path.join(mount_point, f".Trash-{user_id}")
            trash.append((mount_point, path))

        return trash

    def get_info_path(self, path):
        files_dir = os.path.dirname(path)
        trash_dir = os.path.dirname(files_dir)
        info_dir = os.path.join(trash_dir, "info")

        info_path = os.path.join(info_dir, f"{os.path.basename(path)}.trashinfo")
        return info_path

    def get_orig_path(self, path):
        mount_point = utils.find_mount_point(path)
        trash_dir = self._trash.get(mount_point)

        info_path = self.get_info_path(path)
        info_file = ConfigParser(interpolation=None)
        info_file.optionxform = str
        info_file.read(info_path)

        orig_path = unquote(info_file["Trash Info"]["Path"])

        # restore absolute path
        if trash_dir != self._home_trash and not os.path.isabs(orig_path):
            orig_path = os.path.join(mount_point, orig_path)

        return orig_path

    def list(self):
        paths = []

        for name, trash_dir in self._trash.items():
            files_dir = os.path.join(trash_dir, "files")
            if not os.path.exists(files_dir):
                continue
            for filename in os.listdir(files_dir):
                path = os.path.join(files_dir, filename)
                paths.append(path)

        return paths

    def trash(self, path):
        # prevent shutil.move strange behavior
        if not os.access(path, os.W_OK, follow_symlinks=False):
            raise PermissionError(f"Can't access {path}")

        mount_point = utils.find_mount_point(path)

        trash_dir = self._trash.get(mount_point)
        self._setup_trash_dir(trash_dir)

        files_dir = os.path.join(trash_dir, "files")
        info_dir = os.path.join(trash_dir, "info")

        name = utils.find_new_name(files_dir, os.path.basename(path), fmt="%s.%d")
        dest_path = os.path.join(files_dir, name)
        info_path = os.path.join(info_dir, f"{name}.trashinfo")

        if trash_dir == self._home_trash:
            orig_path = path
        else:
            orig_path = os.path.relpath(path, mount_point)

        info_file = ConfigParser(interpolation=None)
        info_file.optionxform = str
        info_file["Trash Info"] = {}
        info_file["Trash Info"]["Path"] = quote(orig_path)
        info_file["Trash Info"]["DeletionDate"] = (
            datetime.utcnow().replace(microsecond=0).isoformat()
        )

        with open(info_path, "w") as info:
            info_file.write(info)
        shutil.move(path, dest_path)

    def restore(self, path):
        orig_path = self.get_orig_path(path)
        info_path = self.get_info_path(path)

        os.remove(info_path)
        shutil.move(path, orig_path)

    def remove(self, path):
        info_path = self.get_info_path(path)
        os.remove(info_path)

        if os.path.isdir(path):
            shutil.rmtree(path)
        else:
            os.remove(path)


default_trash = PortfolioTrash()



# settings.py
#
# Copyright 2021 Martin Abente Lahaye
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GObject


class PortfolioSettings(GObject.GObject):
    __gtype_name__ = "PortfolioSettings"

    SCHEMA = "dev.tchx84.Portfolio"
    ALPHABETICAL_ORDER = "alphabetical"
    MODIFIED_TIME_ORDER = "modified_time"

    def __init__(self):
        super().__init__()
        self._settings = None
        if Gio.SettingsSchemaSource.get_default().lookup(self.SCHEMA, False):
            self._settings = Gio.Settings(self.SCHEMA)

    @GObject.Property(type=bool, default=False)
    def show_hidden(self):
        if self._settings is None:
            return False
        return self._settings.get_boolean("show-hidden")

    @show_hidden.setter
    def show_hidden(self, value):
        if self._settings is None:
            return
        self._settings.set_boolean("show-hidden", value)

    @GObject.Property(type=str)
    def sort_order(self):
        if self._settings is None:
            return self.ALPHABETICAL_ORDER
        return self._settings.get_string("sort-order")

    @sort_order.setter
    def sort_order(self, value):
        if self._settings is None:
            return
        value = [self.ALPHABETICAL_ORDER, self.MODIFIED_TIME_ORDER].index(value)
        self._settings.set_enum("sort-order", value)



# __int__.py
#
# Copyright 2021 Martin Abente Lahaye
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import logging

level = "DEBUG" if os.environ.get("PORTFOLIO_DEBUG") else "INFO"

logger = logging.getLogger("portfolio")
logger.setLevel(level=getattr(logging, level))
logger.addHandler(logging.StreamHandler())



# devices.py
#
# Copyright 2021 Martin Abente Lahaye
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GLib, GObject

from . import logger
from . import utils


class PortfolioDrive(GObject.GObject):
    __gtype_name__ = "PortfolioDrive"

    def __init__(self, object):
        GObject.GObject.__init__(self)

        self._object = object
        self._drive_proxy = object.get_interface("org.freedesktop.UDisks2.Drive")

        self.uuid = self._get_drive_uuid()
        self.is_ejectable = self._get_drive_is_ejectable()
        self.can_power_off = self._get_drive_can_power_off()
        self.model = self._get_drive_model()
        self.size = self._get_drive_size()

    def __repr__(self):
        return f"Drive(uuid={self.uuid}, is_ejectable={self.is_ejectable}, can_power_off={self.can_power_off})"

    def _get_drive_uuid(self):
        return self._drive_proxy.get_cached_property("Id").unpack()

    def _get_drive_is_ejectable(self):
        return self._drive_proxy.get_cached_property("Ejectable").unpack()

    def _get_drive_can_power_off(self):
        return self._drive_proxy.get_cached_property("CanPowerOff").unpack()

    def _get_drive_model(self):
        return self._drive_proxy.get_cached_property("Model").unpack()

    def _get_drive_size(self):
        return self._drive_proxy.get_cached_property("Size").unpack()

    def _on_eject_finished(self, proxy, task, callback, device):
        logger.debug(f"eject finished {self} {device}")
        try:
            proxy.call_finish(task)

            # all devices should power off whenever possible
            if self.can_power_off:
                self.power_off(callback, device)
                return

            device.safely_removed = True
            callback(device, True)
        except Exception as e:
            logger.debug(e)
            callback(device, False)

    def _on_power_off_finished(self, proxy, task, callback, device):
        logger.debug(f"power_off finished {self} {device}")
        try:
            proxy.call_finish(task)
            device.safely_removed = True
            callback(device, True)
        except Exception as e:
            logger.debug(e)
            callback(device, False)

    def eject(self, callback, device):
        logger.debug(f"eject {self} {device}")
        self._drive_proxy.call(
            "Eject",
            GLib.Variant("(a{sv})", ({},)),
            Gio.DBusCallFlags.NONE,
            -1,
            None,
            self._on_eject_finished,
            callback,
            device,
        )

    def power_off(self, callback, device):
        logger.debug(f"power_off {self} {device}")
        self._drive_proxy.call(
            "PowerOff",
            GLib.Variant("(a{sv})", ({},)),
            Gio.DBusCallFlags.NONE,
            -1,
            None,
            self._on_power_off_finished,
            callback,
            device,
        )

    def shutdown(self, callback, device):
        if self.is_ejectable:
            self.eject(callback=callback, device=device)
        elif self.can_power_off:
            self.power_off(callback=callback, device=device)


class PortfolioBlock(GObject.GObject):
    __gtype_name__ = "PortfolioBlock"

    __gsignals__ = {
        "updated": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self, object):
        GObject.GObject.__init__(self)

        self._object = object
        self._block_proxy = object.get_interface("org.freedesktop.UDisks2.Block")

        self.label = self._get_block_label()
        self.uuid = self._get_block_uuid()
        self.drive = self._get_block_drive()
        self.crypto_backing_device = self._get_block_crypto_backing_device()
        self.hint_system = self._get_block_hint_system()
        self.drive_object = None
        self.safely_removed = False

    def __repr__(self):
        return f"Block(uuid={self.uuid}, label={self.label})"

    def _get_block_drive(self):
        return self._block_proxy.get_cached_property("Drive").unpack()

    def _get_block_label(self):
        for property in ["IdLabel", "IdUUID"]:
            if label := self._block_proxy.get_cached_property(property):
                return label.unpack()

        return None

    def _get_block_uuid(self):
        if uuid := self._block_proxy.get_cached_property("IdUUID"):
            return uuid.unpack()

        return None

    def _get_block_crypto_backing_device(self):
        if device := self._block_proxy.get_cached_property("CryptoBackingDevice"):
            return device.unpack()

        return None

    def _get_block_hint_system(self):
        if (hint := self._block_proxy.get_cached_property("HintSystem")) is not None:
            return hint.unpack()

        return None


class PortfolioDevice(PortfolioBlock):
    __gtype_name__ = "PortfolioDevice"

    def __init__(self, object):
        PortfolioBlock.__init__(self, object)

        self._filesystem_proxy = object.get_interface(
            "org.freedesktop.UDisks2.Filesystem"
        )
        self._filesystem_proxy.connect(
            "g-properties-changed", self._on_filesystem_changed
        )

        self.mount_point = self._get_preferred_mount_point()
        self.encrypted_object = None

    def __repr__(self):
        return f"Device(uuid={self.uuid}, label={self.label}, mount_point={self.mount_point})"

    def _get_string_from_bytes(self, bytes):
        return bytearray(bytes).replace(b"\x00", b"").decode("utf-8")

    def _get_filesystem_mount_points(self):
        return [
            self._get_string_from_bytes(m)
            for m in self._filesystem_proxy.get_cached_property("MountPoints")
            if m
        ]

    def _get_fstab_mount_points(self):
        return [
            self._get_string_from_bytes(c.get("dir"))
            for t, c in self._block_proxy.get_cached_property("Configuration")
            if t == "fstab"
        ]

    def _get_preferred_mount_point(self):
        mount_points = self._get_filesystem_mount_points()

        # Check if not mounted yet
        if len(mount_points) == 0:
            return None

        # Check udisks2 to see if there's only one option
        if len(mount_points) == 1:
            return mount_points[0]

        fstab_mount_points = self._get_fstab_mount_points()

        # Filter to active mount points that are in fstab
        fstab_mount_points = list(
            set(fstab_mount_points).intersection(set(mount_points))
        )

        # Check fstab to see if there's a valid option
        if fstab_mount_points:
            return fstab_mount_points[0]

        # Default to previous behavior
        return mount_points[0]

    def _on_filesystem_changed(self, proxy, new_properties, old_properties):
        properties = new_properties.unpack()
        if "MountPoints" in properties:
            self.mount_point = self._get_preferred_mount_point()
            self.emit("updated")

    def _on_mount_finished(self, proxy, task, callback):
        try:
            proxy.call_finish(task)
            callback(self, self.encrypted_object, True)
        except Exception as e:
            logger.debug(e)
            callback(self, self.encrypted_object, False)

    def _on_unmount_finished(self, proxy, task, callback):
        logger.debug(f"unmont finished {self}")
        try:
            proxy.call_finish(task)
        except Exception as e:
            logger.debug(e)
            callback(self, False)
            return

        self._shutdown(callback)

    def _shutdown(self, callback):
        if self.encrypted_object is not None:
            self.encrypted_object.lock(callback, self)
        else:
            self.drive_object.shutdown(callback, self)

    def mount(self, callback):
        self._filesystem_proxy.call(
            "Mount",
            GLib.Variant("(a{sv})", ({},)),
            Gio.DBusCallFlags.NONE,
            -1,
            None,
            self._on_mount_finished,
            callback,
        )

    def unmount(self, callback):
        logger.debug(f"unmont {self}")
        self._filesystem_proxy.call(
            "Unmount",
            GLib.Variant("(a{sv})", ({},)),
            Gio.DBusCallFlags.NONE,
            -1,
            None,
            self._on_unmount_finished,
            callback,
        )

    def eject(self, callback):
        if self.mount_point is not None:
            self.unmount(callback)
        else:
            self._shutdown(callback)


class PortfolioEncrypted(PortfolioBlock):
    __gtype_name__ = "PortfolioEncrypted"

    def __init__(self, object):
        PortfolioBlock.__init__(self, object)

        self._encrypted_proxy = object.get_interface(
            "org.freedesktop.UDisks2.Encrypted"
        )
        self._encrypted_proxy.connect(
            "g-properties-changed", self._on_cleartext_device_changed
        )

        self.mount_point = None
        self.cleartext_device = self._get_encrypted_cleartext_device()
        self.cleartext_device_object = None

    def _get_encrypted_cleartext_device(self):
        return self._encrypted_proxy.get_cached_property("CleartextDevice").unpack()

    def _on_cleartext_device_changed(self, proxy, new_properties, old_properties):
        properties = new_properties.unpack()
        if "CleartextDevice" in properties:
            self.cleartext_device = self._get_encrypted_cleartext_device()
            self.emit("updated")

    def _unlock_finish(self, proxy, task, callback):
        try:
            proxy.call_finish(task)
            self.cleartext_device_object.mount(callback)
        except Exception as e:
            logger.debug(e)
            callback(None, self, False)

    def _lock_finish(self, proxy, task, callback, device):
        logger.debug(f"lock finished {self}")
        try:
            proxy.call_finish(task)
            device.drive_object.shutdown(callback, device)
        except Exception as e:
            logger.debug(e)
            callback(device, False)

    def get_friendly_label(self):
        size = utils.get_size_for_humans(self.drive_object.size)
        return f"{self.drive_object.model} ({size} Drive)"

    def unlock(self, passphrase, callback):
        self._encrypted_proxy.call(
            "Unlock",
            GLib.Variant("(sa{sv})", (passphrase, {})),
            Gio.DBusCallFlags.NONE,
            -1,
            None,
            self._unlock_finish,
            callback,
        )

    def lock(self, callback, device):
        logger.debug(f"lock {self}")
        self._encrypted_proxy.call(
            "Lock",
            GLib.Variant("(a{sv})", ({},)),
            Gio.DBusCallFlags.NONE,
            -1,
            None,
            self._lock_finish,
            callback,
            device,
        )

    def eject(self, callback):
        self.drive_object.shutdown(callback, self)


class PortfolioDevices(GObject.GObject):
    __gtype_name__ = "PortfolioDevices"

    __gsignals__ = {
        "added": (GObject.SignalFlags.RUN_LAST, None, (object,)),
        "removed": (GObject.SignalFlags.RUN_LAST, None, (object,)),
        "encrypted-added": (GObject.SignalFlags.RUN_LAST, None, (object,)),
    }

    def __init__(self):
        GObject.GObject.__init__(self)

        self._drives = {}
        self._devices = {}
        self._encrypted = {}
        self._manager = None

        try:
            self._manager = self._get_manager_proxy()
        except Exception as e:
            logger.debug(f"No udisk2 service found: {e}")
            return

        self._manager.connect("object-added", self._on_object_added)
        self._manager.connect("object-removed", self._on_object_removed)

    def _get_manager_proxy(self):
        return Gio.DBusObjectManagerClient.new_for_bus_sync(
            Gio.BusType.SYSTEM,
            Gio.DBusObjectManagerClientFlags.NONE,
            "org.freedesktop.UDisks2",
            "/org/freedesktop/UDisks2",
            None,
            None,
            None,
        )

    def _on_object_added(self, manager, object):
        self._add_object(object)

    def _on_object_removed(self, manager, object):
        self._remove_object(object)

    def _on_encrypted_updated(self, encrypted):
        if encrypted.cleartext_device in self._devices:
            self.emit("removed", encrypted)

    def _remove_device(self, object_path):
        device = self._devices.get(object_path)
        if device is None:
            return

        self.emit("removed", device)
        del self._devices[object_path]

    def _remove_encrypted(self, object_path):
        encrypted = self._encrypted.get(object_path)
        if encrypted is None:
            return

        # XXX unsafely removed devices don't emit removed signal
        self._remove_device(encrypted.cleartext_device)

        self.emit("removed", encrypted)
        del self._encrypted[object_path]

    def _update_drive_mapping(self):
        # XXX defer added until we can determine real hint system
        for _, device in self._devices.items():
            if device.drive_object is None:
                device.drive_object = self._drives.get(device.drive)
                if device.drive_object and device.hint_system is False:
                    self.emit("added", device)
            if device.drive_object is None:
                if encrypted := self._encrypted.get(device.crypto_backing_device):
                    device.encrypted_object = encrypted
                    encrypted.cleartext_device_object = device
                    device.drive_object = self._drives.get(encrypted.drive)
                if device.drive_object and encrypted.hint_system is False:
                    self.emit("added", device)

        # XXX defer encrypted-added until the corresponding drive is found
        for _, encrypted in self._encrypted.items():
            if encrypted.drive_object is None:
                drive = self._drives.get(encrypted.drive)
                encrypted.drive_object = drive
                if encrypted.hint_system is True:
                    continue
                if drive is None or encrypted.cleartext_device != "/":
                    continue
                self.emit("encrypted-added", encrypted)

    def _add_object(self, object):
        if drive := object.get_interface("org.freedesktop.UDisks2.Drive"):
            self._drives[drive.get_object_path()] = PortfolioDrive(object)
        elif device := object.get_interface("org.freedesktop.UDisks2.Filesystem"):
            self._devices[device.get_object_path()] = PortfolioDevice(object)
        elif proxy := object.get_interface("org.freedesktop.UDisks2.Encrypted"):
            encrypted = PortfolioEncrypted(object)
            encrypted.connect("updated", self._on_encrypted_updated)
            self._encrypted[proxy.get_object_path()] = encrypted

        self._update_drive_mapping()

    def _remove_object(self, object):
        if drive := object.get_interface("org.freedesktop.UDisks2.Drive"):
            del self._drives[drive.get_object_path()]
        elif device := object.get_interface("org.freedesktop.UDisks2.Filesystem"):
            self._remove_device(device.get_object_path())
        elif encrypted := object.get_interface("org.freedesktop.UDisks2.Encrypted"):
            self._remove_encrypted(encrypted.get_object_path())

    def scan(self):
        if self._manager is None:
            return
        for object in self._manager.get_objects():
            self._add_object(object)



# translation.py
#
# Copyright 2021 Clayton Craft
# Copyright 2021 Martin Abente Lahaye
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from . import logger

gettext = lambda msg: msg


def init(localedir):
    global gettext

    try:
        import locale

        gettext = locale.gettext
        locale.bindtextdomain("portfolio", localedir)
        locale.textdomain("portfolio")
    except AttributeError:
        logger.debug("Using fallback gettext module")
        import gettext as _gettext

        gettext = _gettext.gettext
        _gettext.bindtextdomain("portfolio", localedir)
        _gettext.textdomain("portfolio")



# main.py
#
# Copyright 2020 Martin Abente Lahaye
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import gi

gi.require_version("Gtk", "4.0")
gi.require_version("Gio", "2.0")
gi.require_version("Adw", "1")

from gi.repository import Adw, Gio

from .window import PortfolioWindow
from .service import PortfolioService


class Application(Adw.Application):
    def __init__(self):
        super().__init__(
            application_id="dev.tchx84.Portfolio",
            flags=Gio.ApplicationFlags.HANDLES_OPEN,
        )
        self._service = PortfolioService(self)

    def show_properties(self, path):
        self.activate()
        self.props.active_window.show_properties(path, force_page_switch=True)

    def open_path(self, path):
        self.activate()
        self.props.active_window.open(path, force_page_switch=True)

    def do_open(self, files, hint, data):
        self.open_path(files[0].get_path())

    def do_activate(self):
        win = self.props.active_window
        if not win:
            win = PortfolioWindow(application=self)
        win.present()


def main(version):
    app = Application()
    return app.run(sys.argv)



# worker.py
#
# Copyright 2020 Martin Abente Lahaye
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import stat
import time
import shutil
import locale
import datetime
import threading

from pwd import getpwuid
from grp import getgrgid

from gi.repository import Gio, GObject, GLib

from . import utils
from . import logger
from .cache import default_cache
from .translation import gettext as _
from .trash import default_trash


class WorkerStoppedException(Exception):
    pass


class CachedWorker(object):
    def __init__(self):
        default_cache.activate()

    def __del__(self):
        default_cache.deactivate()


class PortfolioWorker(threading.Thread, GObject.GObject):
    __gtype_name__ = "PortfolioWorker"

    def __init__(self):
        threading.Thread.__init__(self)
        GObject.GObject.__init__(self)
        self._cancellable = Gio.Cancellable()

    def _progress(self, current, total):
        logger.debug(current, total)

    def _stop_check(self):
        if self._cancellable.is_cancelled():
            raise WorkerStoppedException()

    def stop(self):
        if not self._cancellable.is_cancelled():
            self._cancellable.cancel()

    def emit(self, *args):
        GLib.idle_add(GObject.GObject.emit, self, *args)


class PortfolioCopyWorker(PortfolioWorker):
    __gtype_name__ = "PortfolioCopyWorker"

    __gsignals__ = {
        "started": (GObject.SignalFlags.RUN_LAST, None, (int,)),
        "updated": (GObject.SignalFlags.RUN_LAST, None, (str, int, int, float, float)),
        "post-update": (GObject.SignalFlags.RUN_LAST, None, (str, str, object, bool)),
        "finished": (GObject.SignalFlags.RUN_LAST, None, (int,)),
        "failed": (GObject.SignalFlags.RUN_LAST, None, (str,)),
        "stopped": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self, selection, directory=None):
        super().__init__()
        self._selection = selection
        self._directory = directory
        self._copy_was_called = False

    def _report_status(self, path, current_bytes, total_bytes):
        self.emit(
            "updated",
            path,
            self._count,
            self._total,
            current_bytes,
            total_bytes,
        )

    def _do_copy(self, source, destination):
        current_bytes = 0
        total_bytes = os.stat(source.name).st_size
        arch_capped = sys.maxsize < 2**32

        # start with an arbitrary block size
        block_min_bytes = 2**23
        block_bytes = block_min_bytes
        last_sync = time.monotonic()

        outfd = destination.fileno()
        infd = source.fileno()

        while True:
            self._stop_check()

            # assume kernel >= 2.6.33
            sent_bytes = os.sendfile(outfd, infd, current_bytes, block_bytes)
            if sent_bytes == 0:
                break

            # fsync frequency is indirectly controlled by the size of the block
            os.fsync(outfd)

            # now dynamically adjust the block size so that it
            # will keep fsync frequency between 750 and 1250ms
            now = time.monotonic()
            if now - last_sync > 1.25:
                block_bytes //= 2
            elif now - last_sync < 0.75:
                block_bytes *= 2

            # on 32 bit arch block size must be capped
            if arch_capped:
                block_bytes = min(block_bytes, 2**30)

            # some lower and upper bounds
            block_bytes = max(block_bytes, block_min_bytes)
            block_bytes = min(block_bytes, total_bytes)

            last_sync = now
            current_bytes += sent_bytes
            self._report_status(destination.name, current_bytes, total_bytes)

        logger.debug(f"sendfile block size converged to {block_bytes}")

    def _copy(self, source_path, destination_path):
        if os.path.islink(source_path):
            os.symlink(os.readlink(source_path), destination_path)
            return

        with open(source_path, "rb") as source:
            with open(destination_path, "wb") as destination:
                self._do_copy(source, destination)

        shutil.copymode(source_path, destination_path)

        self._count += 1
        self._copy_was_called = True

    def run(self):
        self._count = 0
        self._total = sum([utils.count(path) for path, ref in self._selection])

        self.emit("started", self._total)

        for path, ref in self._selection:
            name = os.path.basename(path)
            destination = os.path.join(self._directory, name)
            overwritten = os.path.lexists(destination)

            if path == destination:
                name = utils.find_new_name(self._directory, name)
                destination = os.path.join(self._directory, name)
                overwritten = False

            try:
                self._stop_check()

                if os.path.isdir(path):
                    if overwritten and os.path.isdir(path):
                        shutil.rmtree(destination)
                    shutil.copytree(path, destination, copy_function=self._copy)
                else:
                    self._copy(path, destination)
            except WorkerStoppedException:
                self.emit("stopped")
                return
            except Exception as e:
                logger.debug(e)
                self.emit("failed", destination)
                return
            else:
                utils.sync_folder(os.path.dirname(destination))
                self.emit(
                    "post-update",
                    os.path.basename(destination),
                    destination,
                    utils.get_file_icon(destination),
                    overwritten,
                )

        self.emit("finished", self._total)


class PortfolioCutWorker(PortfolioCopyWorker):
    __gtype_name__ = "PortfolioCutWorker"

    def run(self):
        self._count = 0
        self._total = sum([utils.count(path) for path, ref in self._selection])

        self.emit("started", self._total)

        for path, ref in self._selection:
            name = os.path.basename(path)
            destination = os.path.join(self._directory, name)
            overwritten = os.path.lexists(destination)

            try:
                self._stop_check()

                if destination == path:
                    continue
                if overwritten and os.path.isdir(path):
                    shutil.rmtree(destination)
                if overwritten and os.path.islink(destination):
                    os.unlink(destination)
                shutil.move(path, destination, copy_function=self._copy)
            except WorkerStoppedException:
                self.emit("stopped")
                return
            except Exception as e:
                logger.debug(e)
                self.emit("failed", path)
                return
            else:
                utils.sync_folder(os.path.dirname(destination))

                # XXX force report even if count is not precise
                if self._copy_was_called is False:
                    self._count += 1
                    total_bytes = os.lstat(destination).st_size
                    self._report_status(destination, total_bytes, total_bytes)

                self.emit(
                    "post-update",
                    os.path.basename(destination),
                    destination,
                    utils.get_file_icon(destination),
                    overwritten,
                )

        self.emit("finished", self._total)


class PortfolioDeleteWorker(GObject.GObject, CachedWorker):
    __gtype_name__ = "PortfolioDeleteWorker"

    __gsignals__ = {
        "started": (GObject.SignalFlags.RUN_LAST, None, ()),
        "pre-update": (GObject.SignalFlags.RUN_LAST, None, (str,)),
        "updated": (GObject.SignalFlags.RUN_LAST, None, (str, object, int, int)),
        "finished": (GObject.SignalFlags.RUN_LAST, None, (int,)),
        "failed": (GObject.SignalFlags.RUN_LAST, None, (str,)),
        "stopped": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self, selection):
        super().__init__()
        CachedWorker.__init__(self)
        self._selection = selection
        self._timeout_handler_id = None

    def start(self):
        self.emit("started")

        try:
            paths = []
            for path, ref in self._selection:
                paths += utils.flatten_walk(path)
        except Exception as e:
            logger.debug(e)
            self.emit("failed", path)
            return

        self._paths = paths
        self._refs = dict(self._selection)

        self._total = len(paths)
        self._index = 0
        self._timeout_handler_id = GLib.idle_add(
            self.step, priority=GLib.PRIORITY_HIGH_IDLE + 20
        )

    def step(self):
        if self._index >= self._total:
            self.emit("finished", self._total)
            return

        path = self._paths[self._index]

        try:
            self.emit("pre-update", path)
            if os.path.isdir(path):
                os.rmdir(path)
            else:
                os.unlink(path)
        except Exception as e:
            logger.debug(e)
            self.emit("failed", path)
            return

        self._index += 1
        self.emit("updated", path, self._refs.get(path), self._index, self._total)
        self._timeout_handler_id = GLib.idle_add(
            self.step, priority=GLib.PRIORITY_HIGH_IDLE + 20
        )

    def stop(self):
        if self._timeout_handler_id is not None:
            GLib.Source.remove(self._timeout_handler_id)
            self._timeout_handler_id = None
        self.emit("stopped")


class PortfolioLoadWorker(GObject.GObject, CachedWorker):
    __gtype_name__ = "PortfolioLoadWorker"

    __gsignals__ = {
        "started": (GObject.SignalFlags.RUN_LAST, None, (str,)),
        "updated": (GObject.SignalFlags.RUN_LAST, None, (str, object, int, int)),
        "finished": (GObject.SignalFlags.RUN_LAST, None, (str,)),
        "failed": (GObject.SignalFlags.RUN_LAST, None, (str,)),
    }

    BUFFER = 75

    def __init__(self, directory, hidden=False):
        super().__init__()
        CachedWorker.__init__(self)
        self._directory = directory
        self._hidden = hidden
        self._timeout_handler_id = None

    def start(self):
        self.emit("started", self._directory)

        try:
            self._paths = os.listdir(self._directory)
        except Exception as e:
            logger.debug(e)
            self.emit("failed", self._directory)
            return

        self._total = len(self._paths)
        self._index = 0
        self._timeout_handler_id = GLib.idle_add(
            self.step, priority=GLib.PRIORITY_HIGH_IDLE + 20
        )

    def step(self):
        if self._index >= self._total:
            self.emit("finished", self._directory)
            return

        found = []
        for index in range(0, self.BUFFER):
            if self._index + index < self._total:
                name = self._paths[self._index + index]
                if not self._hidden and name.startswith("."):
                    continue
                path = os.path.join(self._directory, name)
                found.append((name, path, utils.get_file_icon(path)))

        self._index += self.BUFFER
        self.emit("updated", self._directory, found, self._index, self._total)
        self._timeout_handler_id = GLib.idle_add(
            self.step, priority=GLib.PRIORITY_HIGH_IDLE + 20
        )

    def stop(self):
        if self._timeout_handler_id is None:
            return
        GLib.Source.remove(self._timeout_handler_id)
        self._timeout_handler_id = None


class PortfolioOpenWorker(GObject.GObject):
    __gtype_name__ = "PortfolioOpenWorker"

    __gsignals__ = {
        "started": (GObject.SignalFlags.RUN_LAST, None, ()),
        "updated": (GObject.SignalFlags.RUN_LAST, None, ()),
        "finished": (GObject.SignalFlags.RUN_LAST, None, ()),
        "failed": (GObject.SignalFlags.RUN_LAST, None, (str,)),
    }

    def __init__(self, path):
        super().__init__()
        self._path = path
        self._timeout_handler_id = None
        self._is_flatpak = utils.is_flatpak()

    def start(self):
        self.emit("started")

        uri = GLib.Uri.build(
            GLib.UriFlags.NONE, "file", None, None, -1, self._path, None, None
        ).to_string()
        Gio.AppInfo.launch_default_for_uri_async(
            uri, None, None, self._on_launch_finished, None
        )
        self._timeout_handler_id = GLib.timeout_add(100, self._on_step)

    def stop(self):
        pass

    def _on_step(self):
        self.emit("updated")
        return True

    def _on_launch_finished(self, worker, result, data=None):
        if self._timeout_handler_id is not None:
            GLib.Source.remove(self._timeout_handler_id)
            self._timeout_handler_id = None

        if result.had_error():
            self.emit("failed", self._path)
        else:
            self.emit("finished")


class PortfolioPropertiesWorker(GObject.GObject):
    class InnerWorker(threading.Thread):
        def __init__(self, worker):
            super().__init__()
            self._stop = False
            self._worker = worker

        def stop(self):
            self._stop = True

        def run(self):
            size = 0

            for directory, folders, files in os.walk(self._worker._path):
                for filename in files:
                    if self._stop:
                        return
                    try:
                        size += os.lstat(os.path.join(directory, filename)).st_size
                    except:
                        pass

            self._worker._size = utils.get_size_for_humans(size)
            GLib.idle_add(self._worker.notify, "size")

    def __init__(self):
        super().__init__()

        self._path = ""
        self._name = ""
        self._location = ""
        self._type = ""
        self._size = ""
        self._created = ""
        self._modified = ""
        self._accessed = ""
        self._permissions_owner = ""
        self._permissions_group = ""
        self._permissions_others = ""
        self._owner = ""
        self._group = ""

        self._inner_worker = self.InnerWorker(self)

    def _get_file_size(self):
        return os.lstat(self._path).st_size

    def _update_size(self):
        if os.path.isdir(self._path):
            self._inner_worker = self.InnerWorker(self)
            self._inner_worker.start()
        else:
            self._size = utils.get_size_for_humans(self._get_file_size())
            self.notify("size")

    def _get_human_time(self, timestamp):
        fmt = locale.nl_langinfo(locale.D_T_FMT)
        return datetime.datetime.fromtimestamp(timestamp).strftime(fmt)

    def _get_human_permissions(self, path, target="owner"):
        permissions = []
        mode = os.lstat(path).st_mode

        if target == "owner":
            can_read = stat.S_IRUSR
            can_write = stat.S_IWUSR
            can_exec = stat.S_IXUSR
        elif target == "group":
            can_read = stat.S_IRGRP
            can_write = stat.S_IWGRP
            can_exec = stat.S_IXGRP
        else:
            can_read = stat.S_IROTH
            can_write = stat.S_IWOTH
            can_exec = stat.S_IXOTH

        if mode & can_read:
            permissions.append(_("read"))
        if mode & can_write:
            permissions.append(_("write"))
        if mode & can_exec:
            permissions.append(_("execute"))

        if permissions:
            return ", ".join(permissions)

        return _("do nothing")

    def _get_human_owner(self, path):
        owner_id = os.lstat(path).st_uid
        owner = getpwuid(owner_id).pw_name
        return owner

    def _get_human_group(self, path):
        group_id = os.lstat(path).st_gid
        group = getgrgid(group_id).gr_name
        return group

    @GObject.Property(type=str)
    def name(self):
        return self._name

    @GObject.Property(type=str)
    def location(self):
        return self._location

    @GObject.Property(type=str)
    def type(self):
        return self._type

    @GObject.Property(type=str)
    def size(self):
        return self._size

    @GObject.Property(type=str)
    def created(self):
        return self._created

    @GObject.Property(type=str)
    def modified(self):
        return self._modified

    @GObject.Property(type=str)
    def accessed(self):
        return self._accessed

    @GObject.Property(type=str)
    def permissions_owner(self):
        return self._permissions_owner

    @GObject.Property(type=str)
    def permissions_group(self):
        return self._permissions_group

    @GObject.Property(type=str)
    def permissions_others(self):
        return self._permissions_others

    @GObject.Property(type=str)
    def owner(self):
        return self._owner

    @GObject.Property(type=str)
    def group(self):
        return self._group

    @GObject.Property(type=str)
    def path(self):
        return self._path

    @path.setter
    def path(self, path):
        self._path = path

        self._inner_worker.stop()

        file = Gio.File.new_for_path(path)
        info = file.query_info(
            Gio.FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE,
            Gio.FileQueryInfoFlags.NOFOLLOW_SYMLINKS,
            None,
        )

        self._name = os.path.basename(self._path)
        self._location = os.path.dirname(self._path)
        self._type = info.get_content_type()
        self._size = _("Calculating...")
        self._created = self._get_human_time(os.lstat(self._path).st_ctime)
        self._modified = self._get_human_time(os.lstat(self._path).st_mtime)
        self._accessed = self._get_human_time(os.lstat(self._path).st_atime)
        self._permissions_owner = self._get_human_permissions(self._path, "owner")
        self._permissions_group = self._get_human_permissions(self._path, "group")
        self._permissions_others = self._get_human_permissions(self._path, "others")
        self._owner = self._get_human_owner(self._path)
        self._group = self._get_human_group(self._path)

        self.notify("name")
        self.notify("location")
        self.notify("type")
        self.notify("size")
        self.notify("created")
        self.notify("modified")
        self.notify("accessed")
        self.notify("permissions_owner")
        self.notify("permissions_group")
        self.notify("permissions_others")
        self.notify("owner")
        self.notify("group")

        self._update_size()

    def stop(self):
        self._inner_worker.stop()


class PortfolioSendTrashWorker(GObject.GObject, CachedWorker):
    __gtype_name__ = "PortfolioSendTrashWorker"

    __gsignals__ = {
        "started": (GObject.SignalFlags.RUN_LAST, None, ()),
        "pre-update": (GObject.SignalFlags.RUN_LAST, None, (str,)),
        "updated": (GObject.SignalFlags.RUN_LAST, None, (str, object, int, int)),
        "finished": (GObject.SignalFlags.RUN_LAST, None, (int,)),
        "failed": (GObject.SignalFlags.RUN_LAST, None, (str,)),
        "stopped": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self, selection):
        super().__init__()
        CachedWorker.__init__(self)
        self._selection = selection
        self._is_flatpak = utils.is_flatpak()

    def start(self):
        self.emit("started")

        self._paths = [path for path, ref in self._selection]
        self._refs = dict(self._selection)

        self._total = len(self._paths)
        self._index = 0
        self._timeout_handler_id = GLib.idle_add(
            self.step, priority=GLib.PRIORITY_HIGH_IDLE + 20
        )

    def step(self):
        if self._index >= self._total:
            self.emit("finished", self._total)
            return

        path = self._paths[self._index]

        try:
            self.emit("pre-update", path)
            default_trash.trash(path)
        except Exception as e:
            logger.debug(e)
            self.emit("failed", path)
            return

        self._index += 1
        self.emit("updated", path, self._refs.get(path), self._index, self._total)
        self._timeout_handler_id = GLib.idle_add(
            self.step, priority=GLib.PRIORITY_HIGH_IDLE + 20
        )

    def stop(self):
        if self._timeout_handler_id is not None:
            GLib.Source.remove(self._timeout_handler_id)
            self._timeout_handler_id = None
        self.emit("stopped")


class PortfolioRestoreTrashWorker(GObject.GObject, CachedWorker):
    __gtype_name__ = "PortfolioRestoreTrashWorker"

    __gsignals__ = {
        "started": (GObject.SignalFlags.RUN_LAST, None, ()),
        "pre-update": (GObject.SignalFlags.RUN_LAST, None, (str,)),
        "updated": (GObject.SignalFlags.RUN_LAST, None, (str, object, int, int)),
        "finished": (GObject.SignalFlags.RUN_LAST, None, (int,)),
        "failed": (GObject.SignalFlags.RUN_LAST, None, (str,)),
        "stopped": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self, selection):
        super().__init__()
        CachedWorker.__init__(self)
        self._selection = selection

    def start(self):
        self.emit("started")

        self._paths = [path for path, ref in self._selection]
        self._refs = dict(self._selection)

        self._total = len(self._paths)
        self._index = 0
        self._timeout_handler_id = GLib.idle_add(
            self.step, priority=GLib.PRIORITY_HIGH_IDLE + 20
        )

    def step(self):
        if self._index >= self._total:
            self.emit("finished", self._total)
            return

        path = self._paths[self._index]

        try:
            self.emit("pre-update", path)
            default_trash.restore(path)
        except Exception as e:
            logger.debug(e)
            self.emit("failed", path)
            return

        self._index += 1
        self.emit("updated", path, self._refs.get(path), self._index, self._total)
        self._timeout_handler_id = GLib.idle_add(
            self.step, priority=GLib.PRIORITY_HIGH_IDLE + 20
        )

    def stop(self):
        if self._timeout_handler_id is not None:
            GLib.Source.remove(self._timeout_handler_id)
            self._timeout_handler_id = None
        self.emit("stopped")


class PortfolioDeleteTrashWorker(GObject.GObject):
    __gtype_name__ = "PortfolioDeleteTrashWorker"

    __gsignals__ = {
        "started": (GObject.SignalFlags.RUN_LAST, None, ()),
        "pre-update": (GObject.SignalFlags.RUN_LAST, None, (str,)),
        "updated": (GObject.SignalFlags.RUN_LAST, None, (str, object, int, int)),
        "finished": (GObject.SignalFlags.RUN_LAST, None, (int,)),
        "failed": (GObject.SignalFlags.RUN_LAST, None, (str,)),
        "stopped": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self, selection):
        super().__init__()
        self._selection = selection

    def start(self):
        self.emit("started")

        self._paths = [path for path, ref in self._selection]
        self._refs = dict(self._selection)

        self._total = len(self._paths)
        self._index = 0
        self._timeout_handler_id = GLib.idle_add(
            self.step, priority=GLib.PRIORITY_HIGH_IDLE + 20
        )

    def step(self):
        if self._index >= self._total:
            self.emit("finished", self._total)
            return

        path = self._paths[self._index]

        try:
            self.emit("pre-update", path)
            default_trash.remove(path)
        except Exception as e:
            logger.debug(e)
            self.emit("failed", path)
            return

        self._index += 1
        self.emit("updated", path, self._refs.get(path), self._index, self._total)
        self._timeout_handler_id = GLib.idle_add(
            self.step, priority=GLib.PRIORITY_HIGH_IDLE + 20
        )

    def stop(self):
        if self._timeout_handler_id is not None:
            GLib.Source.remove(self._timeout_handler_id)
            self._timeout_handler_id = None
        self.emit("stopped")


class PortfolioLoadTrashWorker(GObject.GObject, CachedWorker):
    __gtype_name__ = "PortfolioLoadTrashWorker"

    __gsignals__ = {
        "started": (GObject.SignalFlags.RUN_LAST, None, (str,)),
        "updated": (GObject.SignalFlags.RUN_LAST, None, (str, object, int, int)),
        "finished": (GObject.SignalFlags.RUN_LAST, None, (str,)),
        "failed": (GObject.SignalFlags.RUN_LAST, None, (str,)),
    }

    def __init__(self, directory=None, hidden=False):
        super().__init__()
        CachedWorker.__init__(self)
        self._timeout_handler_id = None

    def start(self):
        self.emit("started", "")

        try:
            self._paths = default_trash.list()
        except Exception as e:
            logger.debug(e)
            self.emit("failed", "")
            return

        self._total = len(self._paths)
        self._index = 0
        self._timeout_handler_id = GLib.idle_add(
            self.step, priority=GLib.PRIORITY_HIGH_IDLE + 20
        )

    def step(self):
        if self._index >= self._total:
            self.emit("finished", "")
            return

        path = self._paths[self._index]
        name = os.path.basename(path)
        icon = utils.get_file_icon(path)

        self._index += 1
        self.emit("updated", "", [(name, path, icon)], self._index, self._total)
        self._timeout_handler_id = GLib.idle_add(
            self.step, priority=GLib.PRIORITY_HIGH_IDLE + 20
        )

    def stop(self):
        if self._timeout_handler_id is None:
            return
        GLib.Source.remove(self._timeout_handler_id)
        self._timeout_handler_id = None
